#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from timeit import default_timer as timer

import nlopt
import numpy as np
from scipy.optimize import (
    direct as scipy_direct,
    minimize as scipy_minimize,
    minimize_scalar as scipy_minimize_scalar,
)
from opendsm.eemeter.models.daily.optimize_results import OptimizedResult

nlopt_algorithms = {
    "nlopt_direct": nlopt.GN_DIRECT,
    "nlopt_direct_noscal": nlopt.GN_DIRECT_NOSCAL,
    "nlopt_direct_l": nlopt.GN_DIRECT_L,
    "nlopt_direct_l_rand": nlopt.GN_DIRECT_L_RAND,
    "nlopt_direct_l_noscal": nlopt.GN_DIRECT_L_NOSCAL,
    "nlopt_direct_l_rand_noscal": nlopt.GN_DIRECT_L_RAND_NOSCAL,
    "nlopt_orig_direct": nlopt.GN_ORIG_DIRECT,
    "nlopt_orig_direct_l": nlopt.GN_ORIG_DIRECT_L,
    "nlopt_crs2_lm": nlopt.GN_CRS2_LM,
    "nlopt_mlsl_lds": nlopt.G_MLSL_LDS,
    "nlopt_mlsl": nlopt.G_MLSL,
    "nlopt_stogo": nlopt.GD_STOGO,
    "nlopt_stogo_rand": nlopt.GD_STOGO_RAND,
    "nlopt_ags": nlopt.GN_AGS,
    "nlopt_isres": nlopt.GN_ISRES,
    "nlopt_esch": nlopt.GN_ESCH,
    "nlopt_cobyla": nlopt.LN_COBYLA,
    "nlopt_bobyqa": nlopt.LN_BOBYQA,
    "nlopt_newuoa": nlopt.LN_NEWUOA,
    "nlopt_newuoa_bound": nlopt.LN_NEWUOA_BOUND,
    "nlopt_praxis": nlopt.LN_PRAXIS,
    "nlopt_neldermead": nlopt.LN_NELDERMEAD,
    "nlopt_sbplx": nlopt.LN_SBPLX,
    "nlopt_mma": nlopt.LD_MMA,
    "nlopt_ccsaq": nlopt.LD_CCSAQ,
    "nlopt_slsqp": nlopt.LD_SLSQP,
    "nlopt_lbfgs": nlopt.LD_LBFGS,
    "nlopt_tnewton": nlopt.LD_TNEWTON,
    "nlopt_tnewton_precond": nlopt.LD_TNEWTON_PRECOND,
    "nlopt_tnewton_restart": nlopt.LD_TNEWTON_RESTART,
    "nlopt_tnewton_precond_restart": nlopt.LD_TNEWTON_PRECOND_RESTART,
    "nlopt_var1": nlopt.LD_VAR1,
    "nlopt_var2": nlopt.LD_VAR2,
}

nlopt_algorithms = {k.lower(): v for k, v in nlopt_algorithms.items()}

pos_msg = [
    "Optimization terminated successfully.",
    "Optimization terminated: Stop Value was reached.",
    "Optimization terminated: Function tolerance was reached.",
    "Optimization terminated: X tolerance was reached.",
    "Optimization terminated: Max number of evaluations was reached.",
    "Optimization terminated: Max time was reached.",
]
neg_msg = [
    "Optimization failed",
    "Optimization failed: Invalid arguments given",
    "Optimization failed: Out of memory",
    "Optimization failed: Roundoff errors limited progress",
    "Optimization failed: Forced termination",
]


def obj_fcn_dec(obj_fcn, x0, bnds):
    """
    Returns a function that evaluates the objective function with the given bounds.

    Args:
    - obj_fcn: the objective function to be evaluated
    - x0: the initial guess for the optimization
    - bnds: the bounds for the optimization

    Returns:
    - obj_fcn_eval: a function that evaluates the objective function with the given bounds
    - idx_opt: the indices of the variables with non-equal bounds
    """

    idx_opt = [n for n in range(np.shape(bnds)[0]) if (bnds[n, 0] < bnds[n, 1])]

    def obj_fcn_eval(
        x, *args, **kwargs
    ):  # only modify x0 where it has bounds which are not the same
        x0[idx_opt] = x

        return obj_fcn(x0, *args, **kwargs)

    return obj_fcn_eval, idx_opt


class BaseOptimizedResult:
    x = None
    success = None
    status = None
    message = None
    fun = None
    jac = None
    hess = None
    hess_inv = None
    nfev = None
    njev = None
    nhev = None
    nit = None
    maxcv = None
    time_elapsed = None


class BaseOptimizer:
    def __init__(self, obj_fcn, x0, bnds, settings):
        """
        The constructor for the Optimizer class.

        Parameters:
            obj_fcn (function): The objective function to be optimized.
            x0 (np.array): The initial guess for the optimization.
            bnds (list): The bounds for the optimization.
            coef_id (str): The identifier for the coefficient.
            settings (dict): The settings for the optimization.
            opt_settings (Opt_Settings): The settings for the optimization.
        """
        self.bnds = np.array(bnds)
        self.x0 = np.clip(
            x0, bnds[:, 0], bnds[:, 1]
        )  # clip x0 to the bnds, just in case

        self.obj_fcn, self.idx_opt = obj_fcn_dec(obj_fcn, x0, bnds)

        self.settings = settings


class SciPyOptimizer(BaseOptimizer):
    def run(self):
        """
        Optimize the objective function using the SciPy library. Different optimization options are available,
        such as scipy_COBYLA, scipy_SLSQP, scipy_L_BFGS_B, scipy_TNC, scipy_BFGS, scipy_Powell, scipy_Nelder-Mead.
        options argument needs to have the algorithm specified.

        Args:
            x0 (list): Initial guess for the optimization.
            bnds (tuple): Bounds for the optimization.
            settings (Opt_Settings): The settings for the optimization.

        Returns:
            res_out (OptimizedResult): An object containing the results of the optimization.
        """

        settings = self.settings
        x0 = self.x0
        bnds = self.bnds

        timer_start = timer()

        algorithm = settings.algorithm[6:]

        if algorithm.lower() in ["brent", "golden", "bounded"]:
            scipy_obj_fcn = lambda x: self.obj_fcn([x])

            if algorithm.lower() in ["brent", "golden"]:
                res = scipy_minimize_scalar(
                    scipy_obj_fcn, bracket=bnds, method=algorithm.lower()
                )

            elif algorithm.lower() == "bounded":
                res = scipy_minimize_scalar(
                    scipy_obj_fcn, bounds=bnds[0], method="bounded"
                )

            res.x = [res.x]

        else:
            x0_opt = x0[self.idx_opt]
            bnds_opt = bnds[self.idx_opt, :]
            bnds_opt = tuple(map(tuple, bnds_opt))

            scipy_obj_fcn = lambda x: self.obj_fcn(x)

            if algorithm.lower() == "direct":
                res = scipy_direct(
                    scipy_obj_fcn, 
                    bnds_opt,
                    maxiter=int(settings.stop_criteria_value),
                    f_min_rtol=settings.f_tol_rel,
                )
            else:
                res = scipy_minimize(
                    scipy_obj_fcn, x0_opt, method=algorithm, bounds=bnds_opt
                )

        res.time_elapsed = timer() - timer_start

        return res
    

class NLoptOptimizer(BaseOptimizer):
    def run(self):
        """
        Optimize the objective function using the NLopt library.

        Args:
            x0 (ndarray): Initial guess for the optimization.
            bnds (ndarray): Bounds on the variables.
            options (dict): Dictionary of options for the optimization.

        Returns:
            res_out (OptimizedResult): Object containing the results of the optimization.
        """
        settings = self.settings
        x0 = self.x0
        bnds = self.bnds

        timer_start = timer()

        obj_fcn = self.obj_fcn
        idx_opt = self.idx_opt

        x0_opt = x0[idx_opt]
        bnds_opt = bnds[idx_opt, :].T
        
        algorithm = nlopt_algorithms[settings.algorithm]

        opt = nlopt.opt(algorithm, np.size(x0_opt))
        opt.set_min_objective(obj_fcn)
        if settings.stop_criteria_type == "iteration maximum":
            opt.set_maxeval(int(settings.stop_criteria_value) - 1)
        elif settings.stop_criteria_type == "maximum time [min]":
            opt.set_maxtime(settings.stop_criteria_value * 60)

        opt.set_xtol_rel(settings.x_tol_rel)
        opt.set_ftol_rel(settings.f_tol_rel)
        opt.set_lower_bounds(bnds_opt[0])
        opt.set_upper_bounds(bnds_opt[1])

        # initial_step
        max_initial_step = np.max(np.abs(bnds_opt - x0_opt), axis=0)

        initial_step = (bnds_opt[1] - bnds_opt[0]) * settings.initial_step

        # TODO: bring this back in at some point?
        # coef_id_opt = [id for n, id in enumerate(self.coef_id) if n in idx_opt]
        # for n, coef_name in enumerate(coef_id_opt):
        #     if "dd_bp" in coef_name:
        #         initial_step[n] *= 2

        #     if coef_name == "hdd_bp":
        #         initial_step[n] *= -1

        initial_step = np.clip(initial_step, -max_initial_step, max_initial_step)

        x1 = x0_opt + initial_step
        np.putmask(
            initial_step, (x1 < bnds_opt[0]) | (x1 > bnds_opt[1]), -initial_step
        )  # first step in direction of more variable space

        opt.set_initial_step(initial_step)

        # alter default size of population in relevant algorithms
        if settings.algorithm == "nlopt_crs2_lm":
            default_pop_size = 10 * (len(x0_opt) + 1)
        elif settings.algorithm in ["nlopt_mlsl_lds", "nlopt_mlsl"]:
            default_pop_size = 4
        elif settings.algorithm == "nlopt_isres":
            default_pop_size = 20 * (len(x0_opt) + 1)

            opt.set_population(
                int(np.rint(default_pop_size * settings["initial_pop_multiplier"]))
            )

        # if using multistart algorithm as global, set subopt
        if (settings.algorithm == "nlopt_mlsl_lds"):  
            raise NotImplementedError("nlopt_mlsl_lds not implemented")
            local_algorithm = nlopt_algorithms[self.opt_settings.algorithm]
            sub_opt = nlopt.opt(local_algorithm, np.size(x0_opt))
            sub_opt.set_initial_step(initial_step)
            sub_opt.set_xtol_rel(settings.x_tol_rel)
            sub_opt.set_ftol_rel(settings.f_tol_rel)
            opt.set_local_optimizer(sub_opt)

        x_opt = opt.optimize(x0_opt)  # optimize!

        if nlopt.SUCCESS > 0:
            success = True
            msg = pos_msg[nlopt.SUCCESS - 1]
        else:
            success = False
            msg = neg_msg[nlopt.SUCCESS - 1]

        res = BaseOptimizedResult()
        res.x = x_opt
        res.success = success
        res.message = msg
        res.fun = opt.last_optimum_value()
        res.nfev = opt.get_numevals()
        res.time_elapsed = timer() - timer_start

        return res


class InitialGuessOptimizer:
    def __init__(self, obj_fcn, x0, bnds, settings):
        """
        The constructor for the Optimizer class.

        Parameters:
            obj_fcn (function): The objective function to be optimized.
            x0 (np.array): The initial guess for the optimization.
            bnds (list): The bounds for the optimization.
            opt_settings (Opt_Settings): The settings for the optimization.
        """
        self.x0 = np.array(x0)
        self.bnds = np.array(bnds)
        
        self.obj_fcn = obj_fcn

        self.settings = settings

    def run(self):
        """
        This method runs the optimization process.

        Returns:
            OptimizedResult: An object containing the results of the optimization.
        """
        bnds = self.bnds

        res_all = []
        for settings in [self.settings]:
            if len(res_all) == 0:
                x0 = self.x0
            else:
                x0 = res_all[list(res_all.keys())[-1]].x

            if settings.algorithm[:5] == "scipy":
                res = SciPyOptimizer(self.obj_fcn, x0, bnds, settings).run()
            elif settings.algorithm[:5] == "nlopt":
                res = NLoptOptimizer(self.obj_fcn, x0, bnds, settings).run()

            res_all.append(res)

            if (
                settings.algorithm == "nlopt_MLSL_LDS"
            ):  # if using multistart algorithm, break upon finishing loop
                break

        return res_all[-1]


class Optimizer:
    """
    This class is used to perform optimization on a given objective function using either the SciPy or NLopt library.
    The optimization can be performed globally or locally based on the options provided.

    Attributes:
        bnds (np.array): The bounds for the optimization.
        x0 (np.array): The initial guess for the optimization.
        obj_fcn (function): The objective function to be optimized.
        idx_opt (int): The index of the optimal solution.
        coef_id (str): The identifier for the coefficient.
        settings (dict): The settings for the optimization.
        opt_options (dict): The options for the optimization.
    """

    def __init__(self, obj_fcn, x0, bnds, coef_id, settings, opt_settings):
        """
        The constructor for the Optimizer class.

        Parameters:
            obj_fcn (function): The objective function to be optimized.
            x0 (np.array): The initial guess for the optimization.
            bnds (list): The bounds for the optimization.
            coef_id (str): The identifier for the coefficient.
            settings (dict): The settings for the optimization.
            opt_settings (Opt_Settings): The settings for the optimization.
        """
        self.coef_id = coef_id
        self.x0 = np.array(x0)
        self.bnds = np.array(bnds)
        
        self.obj_fcn = obj_fcn

        self.settings = settings
        self.opt_settings = opt_settings

    def run(self):
        """
        This method runs the optimization process.

        Returns:
            OptimizedResult: An object containing the results of the optimization.
        """
        bnds = self.bnds

        res_all = []
        for settings in [self.opt_settings]:
            if len(res_all) == 0:
                x0 = self.x0
            else:
                x0 = res_all[list(res_all.keys())[-1]].x

            if settings.algorithm[:5] == "scipy":
                optimizer_class = SciPyOptimizer
            elif settings.algorithm[:5] == "nlopt":
                optimizer_class = NLoptOptimizer
                
            optimizer = optimizer_class(self.obj_fcn, x0, bnds, settings)
            res = optimizer.run()

            x, mean_loss, TSS, T, model, weight, resid, jac, alpha, C = optimizer.obj_fcn(
                res.x, optimize_flag=False
            )

            res = OptimizedResult(
                x,
                bnds,
                self.coef_id,
                alpha,
                C,
                T,
                model,
                weight,
                resid,
                jac,
                mean_loss,
                TSS,
                res.success,
                res.message,
                res.nfev,
                res.time_elapsed,
                self.settings,
            )

            res_all.append(res)

            if (
                settings.algorithm == "nlopt_MLSL_LDS"
            ):  # if using multistart algorithm, break upon finishing loop
                break

        return res_all[-1]


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numpy as np

from opendsm.common.adaptive_loss import (
    remove_outliers,
    adaptive_weights,
    adaptive_loss_fcn,
)


def test_remove_outliers():
    # Test case 1: No outliers
    data = np.array([1, 2, 3, 4, 5])
    data_no_outliers, idx_no_outliers = remove_outliers(data)
    assert np.array_equal(data, data_no_outliers)
    assert np.array_equal(idx_no_outliers, np.arange(len(data)))

    # Test case 2: Outliers present
    data = np.array([1, 2, 3, 4, 5, 100])
    data_no_outliers, idx_no_outliers = remove_outliers(data)
    assert np.array_equal(data_no_outliers, np.array([1, 2, 3, 4, 5]))
    assert np.array_equal(idx_no_outliers, np.arange(len(data) - 1))

    # Test case 3: Weights provided
    data = np.array([1.0, 2, 3, 4, 5, 100])
    weights = np.array([1, 1, 1, 1, 1, 0.1])
    data_no_outliers, idx_no_outliers = remove_outliers(data, weights)
    assert np.array_equal(data_no_outliers, np.array([1, 2, 3, 4, 5]))
    assert np.array_equal(idx_no_outliers, np.arange(len(data) - 1))


def test_adaptive_loss_fcn():
    # Test case 1: Test with all finite values
    x = np.array([1, 2, 3, 4, 5])
    loss_fcn_val, loss_alpha = adaptive_loss_fcn(x)
    assert np.isfinite(loss_fcn_val)
    assert np.isfinite(loss_alpha)

    # Test case 3: Test with zero values
    x = np.array([0, 0, 0, 0, 0])
    loss_fcn_val, loss_alpha = adaptive_loss_fcn(x)
    assert np.isfinite(loss_fcn_val)
    assert np.isfinite(loss_alpha)

    # Test case 4: Test with negative values
    x = np.array([-1, -2, -3, -4, -5])
    loss_fcn_val, loss_alpha = adaptive_loss_fcn(x)
    assert np.isfinite(loss_fcn_val)
    assert np.isfinite(loss_alpha)

    # Test case 5: Test with large values
    x = np.array([1e10, 2e10, 3e10, 4e10, 5e10])
    loss_fcn_val, loss_alpha = adaptive_loss_fcn(x)
    assert np.isfinite(loss_fcn_val)
    assert np.isfinite(loss_alpha)


def test_adaptive_weights():
    # Test case 1: x has not been standardized
    x = np.array([1, 2, 3, 4, 5])
    weights, C, alpha = adaptive_weights(x)
    assert np.allclose(weights, np.array([1, 1, 1, 0.99993894, 0.99992852]), atol=1e-3)
    assert np.isclose(C, 4.4478)
    assert np.isclose(alpha, 2.0)

    # Test case 2: x has been standardized
    x = np.array([1, 2, 3, 4, 5])
    mu = np.mean(x)
    sigma = np.std(x)
    x = (x - mu) / sigma
    weights, C, alpha = adaptive_weights(x, alpha=3)
    assert np.allclose(weights, np.array([1, 1, 1, 1.02496275, 1.09644634]), atol=1e-3)
    assert np.isclose(C, 3.1450695413615257)
    assert np.isclose(alpha, 3.0)

    # Test case 3: x contains non-finite values
    x = np.array([1, 2, np.nan, 4, 5])
    weights, C, alpha = adaptive_weights(x)
    assert np.allclose(
        weights, np.array([1, 1, 0.99993282, 0.99994308, 0.99993282]), atol=1e-3
    )
    assert np.isclose(C, 5.55975)
    assert np.isclose(alpha, 2.0)

    # Test case 4: x contains outliers
    x = np.array([1, 2, 3, 4, 5, 100])
    weights, C, alpha = adaptive_weights(x)
    assert np.allclose(weights, np.array([1, 1, 1, 0.9865, 0.9479, 0.0011]), atol=1e-3)
    assert np.isclose(C, 6.05975)
    assert np.isclose(alpha, -1.0928, atol=1e-2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from pathlib import Path

import pandas as pd
import requests

from opendsm.common.const import TutorialDataChoice

# Define the current directory
current_dir = Path(__file__).resolve().parent
data_dir = current_dir.parents[1] / "data"

# Set download branch
branch = "master"


comparison_group_time_series = [
    TutorialDataChoice.HOURLY_COMPARISON_GROUP_DATA,
    TutorialDataChoice.DAILY_COMPARISON_GROUP_DATA,
    TutorialDataChoice.MONTHLY_COMPARISON_GROUP_DATA,
]

treatment_time_series = [
    TutorialDataChoice.HOURLY_TREATMENT_DATA,
    TutorialDataChoice.DAILY_TREATMENT_DATA,
    TutorialDataChoice.MONTHLY_TREATMENT_DATA,
]


def load_test_data(data_type: str):
    """Returns back tutorial data of the given data type as a dataframe

    Args:
        data_type (str): Must be one of the following:
            - "features"
            - "seasonal_hourly_day_of_week_loadshape"
            - "seasonal_day_of_week_loadshape"
            - "month_loadshape"
            - "hourly_data"
            - "daily_treatment_data"
            - "monthly_treatment_data"

    Returns:
        (dataframe): Returns a dataframe
    """

    # remove all "_" and " " from string and convert to lowercase
    data_type = data_type.lower()
    data_type = data_type.replace("_", "").replace(" ", "")

    valid_list = [k.value for k in TutorialDataChoice]
    keys = [k.lower() for k in TutorialDataChoice.__members__.keys()]

    if data_type not in valid_list:
        raise ValueError(
            f"Data type {data_type} not recognized. \nMust be one of {keys}."
        )

    if data_type in [*comparison_group_time_series, *treatment_time_series]:
        return _load_time_series_data(data_type)

    else:
        return _load_other_data(data_type)


def download_all_test_data():
    """Downloads all the tutorial data to the data directory"""

    # get all repo files
    repo_full_name = "openeemeter/eemeter"
    path = "data"

    url = f"https://api.github.com/repos/{repo_full_name}/contents/{path}"

    if branch != "master":
        url += f"?ref={branch}"

    r = requests.get(url)
    r.raise_for_status()

    files = [file["name"] for file in r.json() if file["type"] == "file"]

    # download all repo files
    for file in files:
        _download_repo_data_file(file)


def _load_time_series_data(data_type):
    if data_type in comparison_group_time_series:
        df = pd.concat(
            [_load_file("hourly_data_0.parquet"), _load_file("hourly_data_1.parquet")],
            axis=0,
        )

    elif data_type in treatment_time_series:
        df = _load_file("hourly_data_2.parquet")

    # localize datetime and convert to CST
    df = df.reset_index()
    df["datetime"] = df["datetime"].dt.tz_localize("UTC")
    df["datetime"] = df["datetime"] + pd.Timedelta(hours=5)
    df["datetime"] = df["datetime"].dt.tz_convert("America/Chicago")
    df = df.set_index(["id", "datetime"])

    df_baseline = df[["temperature", "observed_baseline"]]
    df_baseline = df_baseline.rename(columns={"observed_baseline": "observed"})

    df_reporting = df[["temperature", "observed_reporting"]]
    df_reporting = df_reporting.rename(columns={"observed_reporting": "observed"})

    df_reporting = df_reporting.reset_index()
    df_reporting["datetime"] = df_reporting["datetime"] + pd.Timedelta(days=365)
    df_reporting = df_reporting.set_index(["id", "datetime"])

    if "daily" in data_type:
        df_baseline = _aggregate_hourly_data(df_baseline, "D")
        df_reporting = _aggregate_hourly_data(df_reporting, "D")

    elif "monthly" in data_type:
        df_baseline = _aggregate_hourly_data(df_baseline, "MS")
        df_reporting = _aggregate_hourly_data(df_reporting, "MS")

    return df_baseline, df_reporting


def _aggregate_hourly_data(df, agg):
    df_agg = df.reset_index().set_index("datetime").groupby("id")
    df_agg_temperature = df_agg["temperature"].resample("D").mean()
    df_agg_observed = df_agg["observed"].resample(agg).sum()

    if agg == "MS":
        df_agg_observed = df_agg_observed.reindex(df_agg_temperature.index)

    df = pd.concat([df_agg_temperature, df_agg_observed], axis=1)
    df = df.reset_index().set_index(["id", "datetime"])

    return df


def _load_other_data(data_type):
    if data_type == TutorialDataChoice.FEATURES:
        df = _load_file("features.csv")

    elif data_type == TutorialDataChoice.SEASONAL_HOUR_DAY_WEEK_LOADSHAPE:
        df = _load_file("seasonal_hourly_day_of_week_loadshape.csv")

    elif data_type == TutorialDataChoice.SEASONAL_DAY_WEEK_LOADSHAPE:
        df = _load_file("seasonal_day_of_week_loadshape.csv")

    elif data_type == TutorialDataChoice.MONTH_LOADSHAPE:
        df = _load_file("month_loadshape.csv")

    df = df.set_index("id")

    return df


def _load_file(file):
    attribution_file = data_dir / "attribution.txt"
    file = data_dir / file
    ext = file.suffix

    # if file does not exist, download it
    if not file.exists():
        # always check for attribution file
        if not attribution_file.exists():
            _download_repo_data_file(attribution_file)

        _download_repo_data_file(file)

    if ext == ".csv":
        df = pd.read_csv(file)

    elif ext == ".parquet":
        df = pd.read_parquet(file)

    else:
        raise ValueError(f"File type {ext} not recognized.")

    return df


def _download_repo_data_file(file: Path):
    repo_full_name = "openeemeter/eemeter"
    path = "data"

    url = f"https://raw.githubusercontent.com/{repo_full_name}/{branch}/{path}/{file.name}"

    r = requests.get(url)
    r.raise_for_status()

    # make directory if it doesn't exist
    if not data_dir.exists():
        data_dir.mkdir()

    with open(data_dir / file, "wb") as f:
        f.write(r.content)


if __name__ == "__main__":
    df = load_test_data("hourly_treatment_data")
    print(df.index.get_level_values(0).nunique())
    print(df.head())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numpy as np
import pandas as pd
import pytest

from opendsm.eemeter.common.features import (
    compute_occupancy_feature,
    compute_temperature_features,
    compute_temperature_bin_features,
    compute_time_features,
    compute_usage_per_day_feature,
    estimate_hour_of_week_occupancy,
    get_missing_hours_of_week_warning,
    fit_temperature_bins,
    merge_features,
)
from opendsm.eemeter.models.hourly_caltrack.segmentation import segment_time_series


def test_compute_temperature_features_no_freq_index(
    il_electricity_cdd_hdd_billing_monthly,
):
    # pick a slice with both hdd and cdd
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    temperature_data.index.freq = None
    with pytest.raises(ValueError):
        compute_temperature_features(meter_data.index, temperature_data)


def test_compute_temperature_features_no_meter_data_tz(
    il_electricity_cdd_hdd_billing_monthly,
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    meter_data.index = meter_data.index.tz_localize(None)
    with pytest.raises(ValueError):
        compute_temperature_features(meter_data.index, temperature_data)


def test_compute_temperature_features_no_temp_data_tz(
    il_electricity_cdd_hdd_billing_monthly,
):
    # pick a slice with both hdd and cdd
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    temperature_data = temperature_data.tz_localize(None)
    with pytest.raises(ValueError):
        compute_temperature_features(meter_data.index, temperature_data)


def test_compute_temperature_features_hourly_temp_mean(il_electricity_cdd_hdd_hourly):
    # pick a slice with both hdd and cdd
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]["2016-03-01":"2016-07-01"]
    temperature_data = il_electricity_cdd_hdd_hourly["temperature_data"][
        "2016-03-01":"2016-07-01"
    ]
    df = compute_temperature_features(meter_data.index, temperature_data)
    assert list(sorted(df.columns)) == [
        "n_hours_dropped",
        "n_hours_kept",
        "temperature_mean",
    ]
    assert df.shape == (2952, 3)

    assert round(df.temperature_mean.mean()) == 62.0


def test_compute_temperature_features_hourly_hourly_degree_days(
    il_electricity_cdd_hdd_hourly, snapshot
):
    # pick a slice with both hdd and cdd
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]["2016-03-01":"2016-07-01"]
    temperature_data = il_electricity_cdd_hdd_hourly["temperature_data"][
        "2016-03-01":"2016-07-01"
    ]
    df = compute_temperature_features(
        meter_data.index,
        temperature_data,
        heating_balance_points=[60, 61],
        cooling_balance_points=[65, 66],
        temperature_mean=False,
        degree_day_method="hourly",
    )
    assert list(sorted(df.columns)) == [
        "cdd_65",
        "cdd_66",
        "hdd_60",
        "hdd_61",
        "n_hours_dropped",
        "n_hours_kept",
    ]
    assert df.shape == (2952, 6)
    snapshot.assert_match(
        [
            round(df.hdd_60.mean(), 2),
            round(df.hdd_61.mean(), 2),
            round(df.cdd_65.mean(), 2),
            round(df.cdd_66.mean(), 2),
            round(df.n_hours_kept.mean(), 2),
            round(df.n_hours_dropped.mean(), 2),
        ],
        "values",
    )


def test_compute_temperature_features_hourly_hourly_degree_days_use_mean_false(
    il_electricity_cdd_hdd_hourly, snapshot
):
    # pick a slice with both hdd and cdd
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]["2016-03-01":"2016-07-01"]
    temperature_data = il_electricity_cdd_hdd_hourly["temperature_data"][
        "2016-03-01":"2016-07-01"
    ]
    df = compute_temperature_features(
        meter_data.index,
        temperature_data,
        heating_balance_points=[60, 61],
        cooling_balance_points=[65, 66],
        temperature_mean=False,
        degree_day_method="hourly",
        use_mean_daily_values=False,
    )
    assert df.shape == (2952, 6)
    assert list(sorted(df.columns)) == [
        "cdd_65",
        "cdd_66",
        "hdd_60",
        "hdd_61",
        "n_hours_dropped",
        "n_hours_kept",
    ]
    snapshot.assert_match(
        [
            round(df.hdd_60.mean(), 2),
            round(df.hdd_61.mean(), 2),
            round(df.cdd_65.mean(), 2),
            round(df.cdd_66.mean(), 2),
            round(df.n_hours_kept.mean(), 2),
            round(df.n_hours_dropped.mean(), 2),
        ],
        "values",
    )


def test_compute_temperature_features_hourly_daily_degree_days_fail(
    il_electricity_cdd_hdd_hourly,
):
    # pick a slice with both hdd and cdd
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]["2016-03-01":"2016-07-01"]
    temperature_data = il_electricity_cdd_hdd_hourly["temperature_data"][
        "2016-03-01":"2016-07-01"
    ]

    with pytest.raises(ValueError):
        compute_temperature_features(
            meter_data.index,
            temperature_data,
            heating_balance_points=[60, 61],
            cooling_balance_points=[65, 66],
            degree_day_method="daily",
        )


def test_compute_temperature_features_hourly_daily_missing_explicit_freq(
    il_electricity_cdd_hdd_hourly,
):
    # pick a slice with both hdd and cdd
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]["2016-03-01":"2016-07-01"]
    temperature_data = il_electricity_cdd_hdd_hourly["temperature_data"][
        "2016-03-01":"2016-07-01"
    ]

    meter_data.index.freq = None
    with pytest.raises(ValueError):
        compute_temperature_features(
            meter_data.index,
            temperature_data,
            heating_balance_points=[60, 61],
            cooling_balance_points=[65, 66],
            degree_day_method="daily",
        )


def test_compute_temperature_features_hourly_bad_degree_days(
    il_electricity_cdd_hdd_hourly,
):
    # pick a slice with both hdd and cdd
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]["2016-03-01":"2016-07-01"]
    temperature_data = il_electricity_cdd_hdd_hourly["temperature_data"][
        "2016-03-01":"2016-07-01"
    ]

    with pytest.raises(ValueError):
        compute_temperature_features(
            meter_data.index,
            temperature_data,
            heating_balance_points=[60, 61],
            cooling_balance_points=[65, 66],
            degree_day_method="UNKNOWN",
        )


def test_compute_temperature_features_hourly_data_quality(
    il_electricity_cdd_hdd_hourly,
):
    # pick a slice with both hdd and cdd
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]["2016-03-01":"2016-07-01"]
    temperature_data = il_electricity_cdd_hdd_hourly["temperature_data"][
        "2016-03-01":"2016-07-01"
    ]

    df = compute_temperature_features(
        meter_data.index, temperature_data, temperature_mean=False, data_quality=True
    )
    assert df.shape == (2952, 4)
    assert list(sorted(df.columns)) == [
        "n_hours_dropped",
        "n_hours_kept",
        "temperature_not_null",
        "temperature_null",
    ]
    assert round(df.temperature_not_null.mean(), 2) == 1.0
    assert round(df.temperature_null.mean(), 2) == 0.0


def test_compute_temperature_features_daily_temp_mean(il_electricity_cdd_hdd_daily):
    meter_data = il_electricity_cdd_hdd_daily["meter_data"]
    temperature_data = il_electricity_cdd_hdd_daily["temperature_data"]
    df = compute_temperature_features(meter_data.index, temperature_data)
    assert df.shape == (810, 3)
    assert list(sorted(df.columns)) == [
        "n_days_dropped",
        "n_days_kept",
        "temperature_mean",
    ]

    assert round(df.temperature_mean.mean()) == 55.0


def test_compute_temperature_features_daily_daily_degree_days(
    il_electricity_cdd_hdd_daily, snapshot
):
    meter_data = il_electricity_cdd_hdd_daily["meter_data"]
    temperature_data = il_electricity_cdd_hdd_daily["temperature_data"]
    df = compute_temperature_features(
        meter_data.index,
        temperature_data,
        heating_balance_points=[60, 61],
        cooling_balance_points=[65, 66],
        temperature_mean=False,
        degree_day_method="daily",
    )
    assert df.shape == (810, 6)
    assert list(sorted(df.columns)) == [
        "cdd_65",
        "cdd_66",
        "hdd_60",
        "hdd_61",
        "n_days_dropped",
        "n_days_kept",
    ]
    snapshot.assert_match(
        [
            round(df.hdd_60.mean(), 2),
            round(df.hdd_61.mean(), 2),
            round(df.cdd_65.mean(), 2),
            round(df.cdd_66.mean(), 2),
            round(df.n_days_kept.mean(), 2),
            round(df.n_days_dropped.mean(), 2),
        ],
        "values",
    )


def test_compute_temperature_features_daily_daily_degree_days_use_mean_false(
    il_electricity_cdd_hdd_daily, snapshot
):
    meter_data = il_electricity_cdd_hdd_daily["meter_data"]
    temperature_data = il_electricity_cdd_hdd_daily["temperature_data"]
    df = compute_temperature_features(
        meter_data.index,
        temperature_data,
        heating_balance_points=[60, 61],
        cooling_balance_points=[65, 66],
        temperature_mean=False,
        degree_day_method="daily",
        use_mean_daily_values=False,
    )
    assert df.shape == (810, 6)
    assert list(sorted(df.columns)) == [
        "cdd_65",
        "cdd_66",
        "hdd_60",
        "hdd_61",
        "n_days_dropped",
        "n_days_kept",
    ]
    snapshot.assert_match(
        [
            round(df.hdd_60.mean(), 2),
            round(df.hdd_61.mean(), 2),
            round(df.cdd_65.mean(), 2),
            round(df.cdd_66.mean(), 2),
            round(df.n_days_kept.mean(), 2),
            round(df.n_days_dropped.mean(), 2),
        ],
        "values",
    )


def test_compute_temperature_features_daily_hourly_degree_days(
    il_electricity_cdd_hdd_daily, snapshot
):
    meter_data = il_electricity_cdd_hdd_daily["meter_data"]
    temperature_data = il_electricity_cdd_hdd_daily["temperature_data"]
    df = compute_temperature_features(
        meter_data.index,
        temperature_data,
        heating_balance_points=[60, 61],
        cooling_balance_points=[65, 66],
        temperature_mean=False,
        degree_day_method="hourly",
    )
    assert df.shape == (810, 6)
    assert list(sorted(df.columns)) == [
        "cdd_65",
        "cdd_66",
        "hdd_60",
        "hdd_61",
        "n_hours_dropped",
        "n_hours_kept",
    ]
    snapshot.assert_match(
        [
            round(df.hdd_60.mean(), 2),
            round(df.hdd_61.mean(), 2),
            round(df.cdd_65.mean(), 2),
            round(df.cdd_66.mean(), 2),
            round(df.n_hours_kept.mean(), 2),
            round(df.n_hours_dropped.mean(), 2),
        ],
        "values",
    )


def test_compute_temperature_features_daily_hourly_degree_days_use_mean_false(
    il_electricity_cdd_hdd_daily, snapshot
):
    meter_data = il_electricity_cdd_hdd_daily["meter_data"]
    temperature_data = il_electricity_cdd_hdd_daily["temperature_data"]
    df = compute_temperature_features(
        meter_data.index,
        temperature_data,
        heating_balance_points=[60, 61],
        cooling_balance_points=[65, 66],
        temperature_mean=False,
        degree_day_method="hourly",
        use_mean_daily_values=False,
    )
    assert df.shape == (810, 6)
    assert list(sorted(df.columns)) == [
        "cdd_65",
        "cdd_66",
        "hdd_60",
        "hdd_61",
        "n_hours_dropped",
        "n_hours_kept",
    ]
    snapshot.assert_match(
        [
            round(df.hdd_60.mean(), 2),
            round(df.hdd_61.mean(), 2),
            round(df.cdd_65.mean(), 2),
            round(df.cdd_66.mean(), 2),
            round(df.n_hours_kept.mean(), 2),
            round(df.n_hours_dropped.mean(), 2),
        ],
        "values",
    )


def test_compute_temperature_features_daily_bad_degree_days(
    il_electricity_cdd_hdd_daily,
):
    meter_data = il_electricity_cdd_hdd_daily["meter_data"]
    temperature_data = il_electricity_cdd_hdd_daily["temperature_data"]
    with pytest.raises(ValueError):
        compute_temperature_features(
            meter_data.index,
            temperature_data,
            heating_balance_points=[60, 61],
            cooling_balance_points=[65, 66],
            degree_day_method="UNKNOWN",
        )


def test_compute_temperature_features_daily_data_quality(il_electricity_cdd_hdd_daily):
    meter_data = il_electricity_cdd_hdd_daily["meter_data"]
    temperature_data = il_electricity_cdd_hdd_daily["temperature_data"]
    df = compute_temperature_features(
        meter_data.index, temperature_data, temperature_mean=False, data_quality=True
    )
    assert df.shape == (810, 4)
    assert list(sorted(df.columns)) == [
        "n_days_dropped",
        "n_days_kept",
        "temperature_not_null",
        "temperature_null",
    ]
    assert round(df.temperature_not_null.mean(), 2) == 23.99
    assert round(df.temperature_null.mean(), 2) == 0.00


def test_compute_temperature_features_billing_monthly_temp_mean(
    il_electricity_cdd_hdd_billing_monthly,
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    df = compute_temperature_features(meter_data.index, temperature_data)
    assert df.shape == (27, 3)
    assert list(sorted(df.columns)) == [
        "n_days_dropped",
        "n_days_kept",
        "temperature_mean",
    ]
    assert round(df.temperature_mean.mean()) == 55.0


def test_compute_temperature_features_billing_monthly_daily_degree_days(
    il_electricity_cdd_hdd_billing_monthly, snapshot
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    df = compute_temperature_features(
        meter_data.index,
        temperature_data,
        heating_balance_points=[60, 61],
        cooling_balance_points=[65, 66],
        temperature_mean=False,
        degree_day_method="daily",
    )
    assert df.shape == (27, 6)
    assert list(sorted(df.columns)) == [
        "cdd_65",
        "cdd_66",
        "hdd_60",
        "hdd_61",
        "n_days_dropped",
        "n_days_kept",
    ]
    snapshot.assert_match(
        [
            round(df.hdd_60.mean(), 2),
            round(df.hdd_61.mean(), 2),
            round(df.cdd_65.mean(), 2),
            round(df.cdd_66.mean(), 2),
            round(df.n_days_kept.mean(), 2),
            round(df.n_days_dropped.mean(), 2),
        ],
        "values",
    )


def test_compute_temperature_features_billing_monthly_daily_degree_days_use_mean_false(
    il_electricity_cdd_hdd_billing_monthly, snapshot
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    df = compute_temperature_features(
        meter_data.index,
        temperature_data,
        heating_balance_points=[60, 61],
        cooling_balance_points=[65, 66],
        temperature_mean=False,
        degree_day_method="daily",
        use_mean_daily_values=False,
    )
    assert df.shape == (27, 6)
    assert list(sorted(df.columns)) == [
        "cdd_65",
        "cdd_66",
        "hdd_60",
        "hdd_61",
        "n_days_dropped",
        "n_days_kept",
    ]
    snapshot.assert_match(
        [
            round(df.hdd_60.mean(), 2),
            round(df.hdd_61.mean(), 2),
            round(df.cdd_65.mean(), 2),
            round(df.cdd_66.mean(), 2),
            round(df.n_days_kept.mean(), 2),
            round(df.n_days_dropped.mean(), 2),
        ],
        "values",
    )


def test_compute_temperature_features_billing_monthly_hourly_degree_days(
    il_electricity_cdd_hdd_billing_monthly, snapshot
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    df = compute_temperature_features(
        meter_data.index,
        temperature_data,
        heating_balance_points=[60, 61],
        cooling_balance_points=[65, 66],
        temperature_mean=False,
        degree_day_method="hourly",
    )
    assert df.shape == (27, 6)
    assert list(sorted(df.columns)) == [
        "cdd_65",
        "cdd_66",
        "hdd_60",
        "hdd_61",
        "n_hours_dropped",
        "n_hours_kept",
    ]
    snapshot.assert_match(
        [
            round(df.hdd_60.mean(), 2),
            round(df.hdd_61.mean(), 2),
            round(df.cdd_65.mean(), 2),
            round(df.cdd_66.mean(), 2),
            round(df.n_hours_kept.mean(), 2),
            round(df.n_hours_dropped.mean(), 2),
        ],
        "values",
    )


def test_compute_temperature_features_billing_monthly_hourly_degree_days_use_mean_false(
    il_electricity_cdd_hdd_billing_monthly, snapshot
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    df = compute_temperature_features(
        meter_data.index,
        temperature_data,
        heating_balance_points=[60, 61],
        cooling_balance_points=[65, 66],
        temperature_mean=False,
        degree_day_method="hourly",
        use_mean_daily_values=False,
    )
    assert df.shape == (27, 6)
    assert list(sorted(df.columns)) == [
        "cdd_65",
        "cdd_66",
        "hdd_60",
        "hdd_61",
        "n_hours_dropped",
        "n_hours_kept",
    ]
    snapshot.assert_match(
        [
            round(df.hdd_60.mean(), 2),
            round(df.hdd_61.mean(), 2),
            round(df.cdd_65.mean(), 2),
            round(df.cdd_66.mean(), 2),
            round(df.n_hours_kept.mean(), 2),
            round(df.n_hours_dropped.mean(), 2),
        ],
        "values",
    )


def test_compute_temperature_features_billing_monthly_bad_degree_day_method(
    il_electricity_cdd_hdd_billing_monthly,
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    with pytest.raises(ValueError):
        compute_temperature_features(
            meter_data.index,
            temperature_data,
            heating_balance_points=[60, 61],
            cooling_balance_points=[65, 66],
            degree_day_method="UNKNOWN",
        )


def test_compute_temperature_features_billing_monthly_data_quality(
    il_electricity_cdd_hdd_billing_monthly,
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    df = compute_temperature_features(
        meter_data.index, temperature_data, temperature_mean=False, data_quality=True
    )
    assert df.shape == (27, 4)
    assert list(sorted(df.columns)) == [
        "n_days_dropped",
        "n_days_kept",
        "temperature_not_null",
        "temperature_null",
    ]
    assert round(df.temperature_not_null.mean(), 2) == 729.23
    assert round(df.temperature_null.mean(), 2) == 0.0


def test_compute_temperature_features_billing_bimonthly_temp_mean(
    il_electricity_cdd_hdd_billing_bimonthly,
):
    meter_data = il_electricity_cdd_hdd_billing_bimonthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_bimonthly["temperature_data"]
    df = compute_temperature_features(meter_data.index, temperature_data)
    assert df.shape == (14, 3)
    assert list(sorted(df.columns)) == [
        "n_days_dropped",
        "n_days_kept",
        "temperature_mean",
    ]
    assert round(df.temperature_mean.mean()) == 55.0


def test_compute_temperature_features_billing_bimonthly_daily_degree_days(
    il_electricity_cdd_hdd_billing_bimonthly, snapshot
):
    meter_data = il_electricity_cdd_hdd_billing_bimonthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_bimonthly["temperature_data"]
    df = compute_temperature_features(
        meter_data.index,
        temperature_data,
        heating_balance_points=[60, 61],
        cooling_balance_points=[65, 66],
        temperature_mean=False,
        degree_day_method="daily",
    )
    assert df.shape == (14, 6)
    assert list(sorted(df.columns)) == [
        "cdd_65",
        "cdd_66",
        "hdd_60",
        "hdd_61",
        "n_days_dropped",
        "n_days_kept",
    ]
    snapshot.assert_match(
        [
            round(df.hdd_60.mean(), 2),
            round(df.hdd_61.mean(), 2),
            round(df.cdd_65.mean(), 2),
            round(df.cdd_66.mean(), 2),
            round(df.n_days_kept.mean(), 2),
            round(df.n_days_dropped.mean(), 2),
        ],
        "values",
    )


def test_compute_temperature_features_billing_bimonthly_hourly_degree_days(
    il_electricity_cdd_hdd_billing_bimonthly, snapshot
):
    meter_data = il_electricity_cdd_hdd_billing_bimonthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_bimonthly["temperature_data"]
    df = compute_temperature_features(
        meter_data.index,
        temperature_data,
        heating_balance_points=[60, 61],
        cooling_balance_points=[65, 66],
        temperature_mean=False,
        degree_day_method="hourly",
    )
    assert df.shape == (14, 6)
    assert list(sorted(df.columns)) == [
        "cdd_65",
        "cdd_66",
        "hdd_60",
        "hdd_61",
        "n_hours_dropped",
        "n_hours_kept",
    ]
    snapshot.assert_match(
        [
            round(df.hdd_60.mean(), 2),
            round(df.hdd_61.mean(), 2),
            round(df.cdd_65.mean(), 2),
            round(df.cdd_66.mean(), 2),
            round(df.n_hours_kept.mean(), 2),
            round(df.n_hours_dropped.mean(), 2),
        ],
        "values",
    )


def test_compute_temperature_features_billing_bimonthly_bad_degree_days(
    il_electricity_cdd_hdd_billing_bimonthly,
):
    meter_data = il_electricity_cdd_hdd_billing_bimonthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_bimonthly["temperature_data"]
    with pytest.raises(ValueError):
        compute_temperature_features(
            meter_data.index,
            temperature_data,
            heating_balance_points=[60, 61],
            cooling_balance_points=[65, 66],
            degree_day_method="UNKNOWN",
        )


def test_compute_temperature_features_billing_bimonthly_data_quality(
    il_electricity_cdd_hdd_billing_bimonthly,
):
    meter_data = il_electricity_cdd_hdd_billing_bimonthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_bimonthly["temperature_data"]
    df = compute_temperature_features(
        meter_data.index, temperature_data, temperature_mean=False, data_quality=True
    )
    assert df.shape == (14, 4)
    assert list(sorted(df.columns)) == [
        "n_days_dropped",
        "n_days_kept",
        "temperature_not_null",
        "temperature_null",
    ]
    assert round(df.temperature_not_null.mean(), 2) == 1478.77
    assert round(df.temperature_null.mean(), 2) == 0.0


def test_compute_temperature_features_shorter_temperature_data(
    il_electricity_cdd_hdd_daily,
):
    meter_data = il_electricity_cdd_hdd_daily["meter_data"]
    temperature_data = il_electricity_cdd_hdd_daily["temperature_data"]

    # drop some data
    temperature_data = temperature_data[:-200]

    df = compute_temperature_features(meter_data.index, temperature_data)
    assert df.shape == (810, 3)
    assert list(sorted(df.columns)) == [
        "n_days_dropped",
        "n_days_kept",
        "temperature_mean",
    ]
    assert round(df.temperature_mean.sum()) == 43958.0


def test_compute_temperature_features_shorter_meter_data(il_electricity_cdd_hdd_daily):
    meter_data = il_electricity_cdd_hdd_daily["meter_data"]
    temperature_data = il_electricity_cdd_hdd_daily["temperature_data"]

    # drop some data
    meter_data = meter_data[:-10]

    df = compute_temperature_features(meter_data.index, temperature_data)
    assert df.shape == (800, 3)
    assert list(sorted(df.columns)) == [
        "n_days_dropped",
        "n_days_kept",
        "temperature_mean",
    ]
    assert round(df.temperature_mean.sum()) == 43904.0
    # ensure last row is NaN'ed
    assert pd.isnull(df.iloc[-1].n_days_kept)


def test_compute_temperature_features_with_duplicated_index(
    il_electricity_cdd_hdd_billing_monthly,
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]

    # these are specifically formed to give a less readable error if
    # duplicates are not caught
    meter_data = pd.concat([meter_data, meter_data]).sort_index()
    temperature_data = temperature_data.iloc[8000:]

    with pytest.raises(ValueError) as excinfo:
        compute_temperature_features(meter_data.index, temperature_data)
    assert str(excinfo.value) == "Duplicates found in input meter trace index."


def test_compute_temperature_features_empty_temperature_data():
    index = pd.DatetimeIndex([], tz="UTC", name="dt", freq="h")
    temperature_data = pd.Series({"value": []}, index=index).astype(float)
    result_index = temperature_data.resample("D").sum().index
    meter_data_hack = pd.DataFrame({"value": 0}, index=result_index)

    with pytest.raises(ValueError):
        df = compute_temperature_features(
            meter_data_hack.index,
            temperature_data,
            heating_balance_points=[65],
            cooling_balance_points=[65],
            degree_day_method="daily",
            use_mean_daily_values=False,
        )


def test_compute_temperature_features_empty_meter_data():
    index = pd.DatetimeIndex([], tz="UTC", name="dt", freq="h")
    temperature_data = pd.Series({"value": 0}, index=index)
    result_index = temperature_data.resample("D").sum().index
    meter_data_hack = pd.DataFrame({"value": []}, index=result_index)
    meter_data_hack.index.freq = None

    with pytest.raises(ValueError):
        df = compute_temperature_features(
            meter_data_hack.index,
            temperature_data,
            heating_balance_points=[65],
            cooling_balance_points=[65],
            degree_day_method="daily",
            use_mean_daily_values=False,
        )


def test_merge_features():
    index = pd.date_range("2017-01-01", periods=100, freq="h", tz="UTC")
    features = merge_features(
        [
            pd.Series(1, index=index, name="a"),
            pd.DataFrame({"b": 2}, index=index),
            pd.DataFrame({"c": 3, "d": 4}, index=index),
        ]
    )
    assert list(features.columns) == ["a", "b", "c", "d"]
    assert features.shape == (100, 4)
    assert features.sum().sum() == 1000
    assert features.a.sum() == 100
    assert features.b.sum() == 200
    assert features.c.sum() == 300
    assert features.d.sum() == 400
    assert features.index[0] == index[0]
    assert features.index[-1] == index[-1]


def test_merge_features_empty_raises():
    with pytest.raises(ValueError):
        features = merge_features([])


@pytest.fixture
def meter_data_hourly():
    index = pd.date_range("2017-01-01", periods=100, freq="h", tz="UTC")
    return pd.DataFrame({"value": 1}, index=index)


def test_compute_usage_per_day_feature_hourly(meter_data_hourly):
    usage_per_day = compute_usage_per_day_feature(meter_data_hourly)
    assert usage_per_day.name == "usage_per_day"
    assert usage_per_day["2017-01-01T00:00:00Z"] == 24
    assert usage_per_day.sum() == 2376.0


def test_compute_usage_per_day_feature_hourly_series_name(meter_data_hourly):
    usage_per_day = compute_usage_per_day_feature(
        meter_data_hourly, series_name="meter_value"
    )
    assert usage_per_day.name == "meter_value"


@pytest.fixture
def meter_data_daily():
    index = pd.date_range("2017-01-01", periods=100, freq="D", tz="UTC")
    return pd.DataFrame({"value": 1}, index=index)


def test_compute_usage_per_day_feature_daily(meter_data_daily):
    usage_per_day = compute_usage_per_day_feature(meter_data_daily)
    assert usage_per_day["2017-01-01T00:00:00Z"] == 1
    assert usage_per_day.sum() == 99.0


@pytest.fixture
def meter_data_billing():
    index = pd.date_range("2017-01-01", periods=100, freq="MS", tz="UTC")
    return pd.DataFrame({"value": 1}, index=index)


def test_compute_usage_per_day_feature_billing(meter_data_billing):
    usage_per_day = compute_usage_per_day_feature(meter_data_billing)
    assert usage_per_day["2017-01-01T00:00:00Z"] == 1.0 / 31
    assert usage_per_day.sum().round(3) == 3.257


@pytest.fixture
def complete_hour_of_week_feature():
    index = pd.date_range("2017-01-01", periods=168, freq="h", tz="UTC")
    time_features = compute_time_features(index, hour_of_week=True)
    hour_of_week_feature = time_features.hour_of_week
    return hour_of_week_feature


def test_get_missing_hours_of_week_warning_ok(complete_hour_of_week_feature):
    warning = get_missing_hours_of_week_warning(complete_hour_of_week_feature)
    assert warning is None


@pytest.fixture
def partial_hour_of_week_feature():
    index = pd.date_range("2017-01-01", periods=84, freq="h", tz="UTC")
    time_features = compute_time_features(index, hour_of_week=True)
    hour_of_week_feature = time_features.hour_of_week
    return hour_of_week_feature


def test_get_missing_hours_of_week_warning_triggered(partial_hour_of_week_feature):
    warning = get_missing_hours_of_week_warning(partial_hour_of_week_feature)
    assert warning.qualified_name is not None
    assert warning.description is not None
    assert warning.data["missing_hours_of_week"] == list(range(60, 144))


def test_compute_time_features_bad_freq():
    index = pd.date_range("2017-01-01", periods=168, freq="D", tz="UTC")
    with pytest.raises(ValueError):
        compute_time_features(index)


def test_compute_time_features_all():
    index = pd.date_range("2017-01-01", periods=168, freq="h", tz="UTC")
    features = compute_time_features(index)
    assert list(features.columns) == ["day_of_week", "hour_of_day", "hour_of_week"]
    assert features.shape == (168, 3)
    assert features.astype(float).sum().sum() == 16464.0
    with pytest.raises(TypeError):  # categoricals
        features.day_of_week.sum()
    with pytest.raises(TypeError):
        features.hour_of_day.sum()
    with pytest.raises(TypeError):
        features.hour_of_week.sum()
    assert features.day_of_week.astype("float").sum() == sum(range(7)) * 24
    assert features.hour_of_day.astype("float").sum() == sum(range(24)) * 7
    assert features.hour_of_week.astype("float").sum() == sum(range(168))
    assert features.index[0] == index[0]
    assert features.index[-1] == index[-1]


def test_compute_time_features_none():
    index = pd.date_range("2017-01-01", periods=168, freq="h", tz="UTC")
    with pytest.raises(ValueError):
        compute_time_features(
            index, hour_of_week=False, day_of_week=False, hour_of_day=False
        )


@pytest.fixture
def occupancy_precursor(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_hourly["temperature_data"]
    time_features = compute_time_features(meter_data.index)
    temperature_features = compute_temperature_features(
        meter_data.index,
        temperature_data,
        heating_balance_points=[50],
        cooling_balance_points=[65],
        degree_day_method="hourly",
    )
    return merge_features(
        [meter_data.value.to_frame("meter_value"), temperature_features, time_features]
    )


def test_estimate_hour_of_week_occupancy_no_segmentation(occupancy_precursor):
    occupancy = estimate_hour_of_week_occupancy(occupancy_precursor)
    assert list(occupancy.columns) == ["occupancy"]
    assert occupancy.shape == (168, 1)
    assert occupancy.sum().sum() == 0


@pytest.fixture
def one_month_segmentation(occupancy_precursor):
    return segment_time_series(occupancy_precursor.index, segment_type="one_month")


def test_estimate_hour_of_week_occupancy_one_month_segmentation(
    occupancy_precursor, one_month_segmentation
):
    occupancy = estimate_hour_of_week_occupancy(
        occupancy_precursor, segmentation=one_month_segmentation
    )
    assert list(occupancy.columns) == [
        "jan",
        "feb",
        "mar",
        "apr",
        "may",
        "jun",
        "jul",
        "aug",
        "sep",
        "oct",
        "nov",
        "dec",
    ]
    assert occupancy.shape == (168, 12)
    assert occupancy.sum().sum() == 84.0


@pytest.fixture
def temperature_means():
    index = pd.date_range("2017-01-01", periods=2000, freq="h", tz="UTC")
    return pd.DataFrame({"temperature_mean": [10, 35, 55, 80, 100] * 400}, index=index)


def test_fit_temperature_bins_no_segmentation(temperature_means):
    bins = fit_temperature_bins(
        temperature_means, segmentation=None, occupancy_lookup=None
    )
    assert list(bins.columns) == ["keep_bin_endpoint"]
    assert bins.shape == (6, 1)
    assert bins.sum().sum() == 4


@pytest.fixture
def occupancy_lookup_no_segmentation(occupancy_precursor):
    occupancy = estimate_hour_of_week_occupancy(occupancy_precursor)
    return occupancy


def test_fit_temperature_bins_no_segmentation_with_occupancy(
    temperature_means, occupancy_lookup_no_segmentation
):
    occupied_bins, unoccupied_bins = fit_temperature_bins(
        temperature_means,
        segmentation=None,
        occupancy_lookup=occupancy_lookup_no_segmentation,
    )
    assert list(occupied_bins.columns) == ["keep_bin_endpoint"]
    assert occupied_bins.shape == (6, 1)
    assert occupied_bins.sum().sum() == 0

    assert list(unoccupied_bins.columns) == ["keep_bin_endpoint"]
    assert unoccupied_bins.shape == (6, 1)
    assert unoccupied_bins.sum().sum() == 4


def test_fit_temperature_bins_one_month_segmentation(
    temperature_means, one_month_segmentation
):
    bins = fit_temperature_bins(temperature_means, segmentation=one_month_segmentation)
    assert list(bins.columns) == [
        "jan",
        "feb",
        "mar",
        "apr",
        "may",
        "jun",
        "jul",
        "aug",
        "sep",
        "oct",
        "nov",
        "dec",
    ]
    assert bins.shape == (6, 12)
    assert bins.sum().sum() == 12


@pytest.fixture
def occupancy_lookup_one_month_segmentation(
    occupancy_precursor, one_month_segmentation
):
    occupancy_lookup = estimate_hour_of_week_occupancy(
        occupancy_precursor, segmentation=one_month_segmentation
    )
    return occupancy_lookup


def test_fit_temperature_bins_with_occupancy_lookup(
    temperature_means, one_month_segmentation, occupancy_lookup_one_month_segmentation
):
    occupied_bins, unoccupied_bins = fit_temperature_bins(
        temperature_means,
        segmentation=one_month_segmentation,
        occupancy_lookup=occupancy_lookup_one_month_segmentation,
    )
    assert list(occupied_bins.columns) == [
        "jan",
        "feb",
        "mar",
        "apr",
        "may",
        "jun",
        "jul",
        "aug",
        "sep",
        "oct",
        "nov",
        "dec",
    ]
    assert occupied_bins.shape == (6, 12)
    assert occupied_bins.sum().sum() == 0

    assert list(unoccupied_bins.columns) == [
        "jan",
        "feb",
        "mar",
        "apr",
        "may",
        "jun",
        "jul",
        "aug",
        "sep",
        "oct",
        "nov",
        "dec",
    ]
    assert unoccupied_bins.shape == (6, 12)
    assert unoccupied_bins.sum().sum() == 12


def test_fit_temperature_bins_empty(temperature_means):
    bins = fit_temperature_bins(temperature_means.iloc[:0])
    assert list(bins.columns) == ["keep_bin_endpoint"]
    assert bins.shape == (6, 1)
    assert bins.sum().sum() == 0


def test_compute_temperature_bin_features(temperature_means):
    temps = temperature_means.temperature_mean
    bin_features = compute_temperature_bin_features(temps, [25, 75])
    assert list(bin_features.columns) == ["bin_0", "bin_1", "bin_2"]
    assert bin_features.shape == (2000, 3)
    assert bin_features.sum().sum() == 112000.0


@pytest.fixture
def even_occupancy():
    return pd.Series([i % 2 == 0 for i in range(168)], index=pd.Categorical(range(168)))


def test_compute_occupancy_feature(even_occupancy):
    index = pd.date_range("2017-01-01", periods=1000, freq="h", tz="UTC")
    time_features = compute_time_features(index, hour_of_week=True)
    hour_of_week = time_features.hour_of_week
    occupancy = compute_occupancy_feature(hour_of_week, even_occupancy)
    assert occupancy.name == "occupancy"
    assert occupancy.shape == (1000,)
    assert occupancy.sum().sum() == 500


def test_compute_occupancy_feature_with_nans(even_occupancy):
    """If there are less than 168 periods, the NaN at the end causes problems"""
    index = pd.date_range("2017-01-01", periods=100, freq="h", tz="UTC")
    time_features = compute_time_features(index, hour_of_week=True)
    time_features.iloc[-1, time_features.columns.get_loc("hour_of_week")] = np.nan
    hour_of_week = time_features.hour_of_week
    #  comment out line below to see the error from not dropping na when
    # calculationg _add_weights when there are less than 168 periods.

    # TODO (ssuffian): Refactor so get_missing_hours_warnings propogates.
    # right now, it will error if the dropna below isn't used.
    hour_of_week.dropna(inplace=True)
    occupancy = compute_occupancy_feature(hour_of_week, even_occupancy)
    assert occupancy.name == "occupancy"
    assert occupancy.shape == (99,)
    assert occupancy.sum().sum() == 50


@pytest.fixture
def occupancy_precursor_only_nan(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    meter_data = meter_data["2017-01-04":"2017-06-01"].copy()
    meter_data.iloc[-1] = np.nan
    # Simulates a segment where there is only a single nan value
    temperature_data = il_electricity_cdd_hdd_hourly["temperature_data"]
    time_features = compute_time_features(meter_data.index)
    temperature_features = compute_temperature_features(
        meter_data.index,
        temperature_data,
        heating_balance_points=[50],
        cooling_balance_points=[65],
        degree_day_method="hourly",
    )
    return merge_features(
        [meter_data.value.to_frame("meter_value"), temperature_features, time_features]
    )


@pytest.fixture
def segmentation_only_nan(occupancy_precursor_only_nan):
    return segment_time_series(
        occupancy_precursor_only_nan.index, segment_type="three_month_weighted"
    )


def test_estimate_hour_of_week_occupancy_segmentation_only_nan(
    occupancy_precursor_only_nan, segmentation_only_nan
):
    occupancy = estimate_hour_of_week_occupancy(
        occupancy_precursor_only_nan, segmentation=segmentation_only_nan
    )


def test_compute_occupancy_feature_hour_of_week_has_nan(even_occupancy):
    index = pd.date_range("2017-01-01", periods=72, freq="h", tz="UTC")
    time_features = compute_time_features(index, hour_of_week=True)
    hour_of_week = time_features.hour_of_week
    hour_of_week.iloc[-1] = np.nan
    occupancy = compute_occupancy_feature(hour_of_week, even_occupancy)
    assert occupancy.name == "occupancy"
    assert occupancy.shape == (72,)
    assert occupancy.sum() == 36



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import numpy as np

from opendsm.common.utils import OoM
from opendsm.common.utils import fast_std as stdev


def get_idx(A, B):
    """
    Returns a sorted list of indices of items in A that are found in any string in B.

    Parameters:
    A (list): List of items to search for in B.
    B (list): List of strings to search for items in A.

    Returns:
    list: Sorted list of indices of items in A that are found in any string in B.
    """

    idx = []
    for item in A:
        try:
            idx.extend([B.index(txt) for txt in B if item in txt])
        except:
            continue

    idx.sort()

    return idx


def no_weights_obj_fcn(X, aux_inputs):
    """
    Calculates the sum of squared errors (SSE) between the model output and the observed data.

    Parameters:
        X (array-like): Input values for the model.
        aux_inputs (tuple): A tuple containing the model function, observed data, and breakpoint indices.

    Returns:
        float: The SSE between the model output and the observed data.
    """

    model_fcn, obs, idx_bp = aux_inputs

    # flip breakpoints if they are not in the correct order
    if (len(idx_bp) > 1) and (X[idx_bp[0]] > X[idx_bp[1]]):
        X[idx_bp[1]], X[idx_bp[0]] = X[idx_bp[0]], X[idx_bp[1]]

    model = model_fcn(X)
    resid = model - obs

    SSE = np.sum(resid**2)

    return SSE


def model_fcn_dec(model_fcn_full, T_fit_bnds, T):
    """
    Returns a function that takes only X as input and returns the output of model_fcn_full with X, T_fit_bnds, and T as inputs.

    Parameters:
    - model_fcn_full: function
        The full model function that takes X, T_fit_bnds, and T as inputs.
    - T_fit_bnds: tuple
        The bounds of the temperature range.
    - T: float
        The temperature value.

    Returns:
    - function
        A function that takes only X as input and returns the output of model_fcn_full with X, T_fit_bnds, and T as inputs.
    """

    def model_fcn_X_only(X):
        return model_fcn_full(*X, T_fit_bnds, T)

    return model_fcn_X_only


def obj_fcn_decorator(
    model_fcn_full,
    weight_fcn,
    TSS_fcn,
    T,
    obs,
    base_weights,
    settings,
    alpha=2.0,
    coef_id=[],
    initial_fit=True,
):
    """
    A decorator function that calculates the elastic net penalty for a given set of inputs and the objective function
    for input to optimization algorithms.

    Parameters:
    model_fcn_full (function): The full model function.
    weight_fcn (function): The weight function.
    TSS_fcn (function): The TSS (Total Sum of Squares) function.
    T (array-like): The temperature array.
    obs (array-like): The observation array.
    settings (object): The DailySettings object.
    alpha (float): The alpha value for the elastic net penalty. Default is 2.0.
    coef_id (list): The list of coefficient IDs. Default is an empty list.
    initial_fit (bool): Whether or not this is the initial fit. Default is True.

    Returns:
    obj_fcn (function): an objective function having the required inputs for optimization via SciPy and NLopt.
    """

    N = np.shape(obs)[0]
    N_min = settings.segment_minimum_count  # N minimum for a sloped segment
    sigma = 2.698  # 1.5 IQR
    quantile = 0.25
    min_weight = 0.00

    T_fit_bnds = np.array([np.min(T), np.max(T)])
    T_range = T_fit_bnds[1] - T_fit_bnds[0]

    model_fcn = model_fcn_dec(model_fcn_full, T_fit_bnds, T)

    lasso_a = settings.regularization_percent_lasso * settings.regularization_alpha
    ridge_a = (
        1 - settings.regularization_percent_lasso
    ) * settings.regularization_alpha

    idx_k = get_idx(["dd_k"], coef_id)
    idx_beta = get_idx(["dd_beta"], coef_id)
    idx_bp = get_idx(["dd_bp"], coef_id)
    # idx_reg = get_idx(["dd_beta", "dd_k"], coef_id) # drop bps and intercept from regularization
    idx_reg = get_idx(
        ["dd_beta", "dd_k", "dd_bp"], coef_id
    )  # drop intercept from regularization

    def elastic_net_penalty(X, T_sorted, obs_sorted, weight_sorted, wRMSE):
        """
        Calculates the elastic net penalty for a given set of inputs. The elastic net is a regularized
        regression method that linearly combines the L1 and L2 penalties of the lasso and ridge methods.

        Parameters:
        X (array-like): The input array.
        T_sorted (array-like): The sorted temperature array.
        obs_sorted (array-like): The sorted observation array.
        weight_sorted (array-like): The sorted weight array.
        wRMSE (float): The weighted root mean squared error.

        Returns:
        penalty (float): The elastic net penalty.
        """

        # Elastic net
        X_enet = np.array(X).copy()

        ## Scale break points ##
        if len(idx_bp) > 0:
            X_enet[idx_bp] = [
                np.min(np.abs(X_enet[idx] - T_fit_bnds)) for idx in idx_bp
            ]

            if len(idx_bp) == 2:
                X_enet[idx_bp] += (X[idx_bp][1] - X[idx_bp][0]) / 2

            X_enet[idx_bp] *= wRMSE / T_range

        # Find idx for regions
        if len(idx_bp) == 2:
            [hdd_bp, cdd_bp] = X[idx_bp]

            idx_hdd = np.argwhere(T_sorted < hdd_bp).flatten()
            idx_tidd = np.argwhere(
                (hdd_bp <= T_sorted) & (T_sorted <= cdd_bp)
            ).flatten()
            idx_cdd = np.argwhere(cdd_bp < T_sorted).flatten()

        elif len(idx_bp) == 1:
            bp = X[idx_bp]
            if X_enet[idx_beta] < 0:  # HDD_TIDD
                idx_hdd = np.argwhere(T_sorted <= bp).flatten()
                idx_tidd = np.argwhere(bp < T_sorted).flatten()
                idx_cdd = np.array([])

            else:
                idx_hdd = np.array([])  # CDD_TIDD
                idx_tidd = np.argwhere(T_sorted < bp).flatten()
                idx_cdd = np.argwhere(bp <= T_sorted).flatten()

        else:
            idx_hdd = np.array([])
            idx_tidd = np.arange(0, len(T_sorted))
            idx_cdd = np.array([])

        len_hdd = len(idx_hdd)
        len_tidd = len(idx_tidd)
        len_cdd = len(idx_cdd)

        # combine tidd with hdd/cdd if cdd/hdd are large enough to get stdev
        if (len_hdd < N_min) and (len_cdd >= N_min):
            idx_hdd = np.hstack([idx_hdd, idx_tidd])
        elif (len_hdd >= N_min) and (len_cdd < N_min):
            idx_cdd = np.hstack([idx_tidd, idx_cdd])

        # change to idx_hdd and idx_cdd to int arrays
        idx_hdd = idx_hdd.astype(int)
        idx_cdd = idx_cdd.astype(int)

        ## Normalize slopes ##
        # calculate stdevs
        if (len(idx_bp) == 2) and (len(idx_hdd) >= N_min) and (len(idx_cdd) >= N_min):
            N_beta = np.array([len_hdd, len_cdd])
            T_stdev = np.array(
                [
                    stdev(T_sorted[idx_hdd], weights=weight_sorted[idx_hdd]),
                    stdev(T_sorted[idx_cdd], weights=weight_sorted[idx_cdd]),
                ]
            )
            obs_stdev = np.array(
                [
                    stdev(obs_sorted[idx_hdd], weights=weight_sorted[idx_hdd]),
                    stdev(obs_sorted[idx_cdd], weights=weight_sorted[idx_cdd]),
                ]
            )

        elif (len(idx_bp) == 1) and (len(idx_hdd) >= N_min):
            N_beta = np.array([len_hdd])
            T_stdev = stdev(T_sorted[idx_hdd], weights=weight_sorted[idx_hdd])
            obs_stdev = stdev(obs_sorted[idx_hdd], weights=weight_sorted[idx_hdd])

        elif (len(idx_bp) == 1) and (len(idx_cdd) >= N_min):
            N_beta = np.array([len_cdd])
            T_stdev = stdev(T_sorted[idx_cdd], weights=weight_sorted[idx_cdd])
            obs_stdev = stdev(obs_sorted[idx_cdd], weights=weight_sorted[idx_cdd])

        else:
            N_beta = np.array([len_tidd])
            T_stdev = stdev(T_sorted, weights=weight_sorted)
            obs_stdev = stdev(obs_sorted, weights=weight_sorted)

        X_enet[idx_beta] *= T_stdev / obs_stdev

        # add penalty to slope for not having enough datapoints
        X_enet[idx_beta] = np.where(
            N_beta < N_min, X_enet[idx_beta] * 1e30, X_enet[idx_beta]
        )

        ## Scale smoothing parameter ##
        if len(idx_k) > 0:  # reducing X_enet size allows for more smoothing
            X_enet[idx_k] = X[idx_k]

            if (len(idx_k) == 2) and (np.sum(X_enet[idx_k]) > 1):
                X_enet[idx_k] /= np.sum(X_enet[idx_k])

            X_enet[idx_k] *= (
                X_enet[idx_beta] / 2
            )  # uncertain what to divide by, this seems to work well

        X_enet = X_enet[idx_reg]

        if ridge_a == 0:
            penalty = lasso_a * np.linalg.norm(X_enet, 1)
        else:
            penalty = lasso_a * np.linalg.norm(X_enet, 1) + ridge_a * np.linalg.norm(
                X_enet, 2
            )

        return penalty

    def obj_fcn(X, grad=[], optimize_flag=True):
        """
        Creates an objective function having the required inputs for optimization via SciPy and NLopt. If the optimize_flag is true,
        only return the loss. If the optimize_flag is false, return the loss and the model output parameters.

        Parameters:
        - X: array-like
            Array of coefficients.
        - grad: array-like, optional
            Gradient array. Default is an empty list.
        - optimize_flag: bool, optional
            Whether to optimize. Default is True.

        Returns:
        - obj: float
            Objective function value.
        """
        X = np.array(X)

        model = model_fcn(X)
        idx_sorted = np.argsort(T).flatten()
        idx_initial = np.argsort(idx_sorted).flatten()
        resid = model - obs

        T_sorted = T[idx_sorted]
        # model_sorted = model[idx_sorted]
        obs_sorted = obs[idx_sorted]
        resid_sorted = resid[idx_sorted]

        weight_sorted, c, a = weight_fcn(
            *X, T_sorted, resid_sorted, sigma, quantile, alpha, min_weight
        )
        if base_weights is not None:
            weight_sorted *= base_weights[idx_sorted]
            weight_sorted /= np.sum(weight_sorted)

        weight = weight_sorted[idx_initial]
        wSSE = np.sum(weight * resid**2)
        loss = wSSE / N

        if settings.regularization_alpha != 0:
            loss += elastic_net_penalty(
                X, T_sorted, obs_sorted, weight_sorted, np.sqrt(loss)
            )

        if optimize_flag:
            return loss

        else:
            if ("r_squared" in settings.split_selection.criteria) and callable(TSS_fcn):
                TSS = TSS_fcn(*X, T_sorted, obs_sorted)
            else:
                TSS = wSSE

            if initial_fit:
                jac = None
            else:
                eps = 10 ** (OoM(X, method="floor") - 2)
                X_lower = X - eps
                X_upper = X + eps

                # select correct finite difference scheme based on variable type and value
                # NOTE: finite differencing was not returning great results. Looking into switching to JAX autodiff
                fd_type = ["central"] * len(X)
                for i in range(len(X)):
                    if i in idx_k:
                        if X_lower[i] < 0:
                            fd_type[i] = "forward"
                        elif X_upper[i] > 1:
                            fd_type[i] = "backward"
                    elif i in idx_beta:
                        if (X[i] > 0) and (X_lower[i] < 0):
                            fd_type[i] = "forward"
                        elif (X[i] < 0) and (X_upper[i] > 0):
                            fd_type[i] = "backward"
                    elif i in idx_bp:
                        if X_lower[i] < T_sorted[0]:
                            fd_type[i] = "forward"
                        elif X_lower[i] > T_sorted[-1]:
                            fd_type[i] = "backward"

                # https://stackoverflow.com/questions/70572362/compute-efficiently-hessian-matrices-in-jax
                # hess = jit(jacfwd(jacrev(no_weights_obj_fcn), has_aux=True), has_aux=True)(X, [model_fcn, obs, idx_bp])
                # print(hess)
                # obj_grad_fcn = lambda X: no_weights_obj_fcn(X, [model_fcn, obs, idx_bp])

                # jac = numerical_jacobian(obj_grad_fcn, X, dx=eps, fd_type=fd_type)
                jac = None

            return X, loss, TSS, T, model, weight, resid, jac, np.mean(a), c

    return obj_fcn



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""A module for assiting with input/output operations.

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from __future__ import annotations

import datetime

import numpy as np
import pandas as pd
from pandas._typing import (
    CompressionOptions,
    CSVEngine,
    DtypeArg,
    DtypeBackend,
    FilePath,
    IndexLabel,
    ReadCsvBuffer,
    StorageOptions,
    WriteBuffer,
)

__all__ = (
    "meter_data_from_csv",
    "meter_data_from_json",
    "meter_data_to_csv",
    "temperature_data_from_csv",
    "temperature_data_from_json",
    "temperature_data_to_csv",
)


def meter_data_from_csv(
    filepath_or_buffer: str | FilePath | ReadCsvBuffer[bytes] | ReadCsvBuffer[str],
    tz: str | datetime.tzinfo | None = None,
    start_col: str = "start",
    value_col: str = "value",
    gzipped: bool = False,
    freq: str | None = None,
    **kwargs,
) -> pd.DataFrame:
    """Load meter data from a CSV file and convert to a dataframe.

    Note: This is an example of the default csv structure assumed.
        ```python
        start,value
        2017-01-01T00:00:00+00:00,0.31
        2017-01-02T00:00:00+00:00,0.4
        2017-01-03T00:00:00+00:00,0.58
        ```

    Args:
        filepath_or_buffer: File path or object.
        tz: Timezone represented in the meter data. Ex: `UTC` or `US/Pacific`
        start_col: Date period start column.
        value_col: Value column, can be in any unit.
        gzipped: Whether file is gzipped.
        freq: If given, apply frequency to data using `pandas.DataFrame.resample`. One of `['hourly', 'daily']`.
        **kwargs: Extra keyword arguments to pass to `pandas.read_csv`, such as `sep='|'`.
    """

    read_csv_kwargs = {
        "usecols": [start_col, value_col],
        "dtype": {value_col: np.float64},
        "parse_dates": [start_col],
        "index_col": start_col,
    }

    if gzipped:
        read_csv_kwargs.update({"compression": "gzip"})

    # allow passing extra kwargs
    read_csv_kwargs.update(kwargs)

    df = pd.read_csv(filepath_or_buffer, **read_csv_kwargs)
    df.index = pd.to_datetime(df.index, utc=True)

    # for pandas<0.24, which doesn't localize even with utc=True
    if df.index.tz is None:
        df.index = df.index.tz_localize("UTC")  # pragma: no cover

    if tz is not None:
        df = df.tz_convert(tz)

    if freq == "hourly":
        df = df.resample("h").sum(min_count=1)
    elif freq == "daily":
        df = df.resample("D").sum(min_count=1)

    return df


def temperature_data_from_csv(
    filepath_or_buffer: str | FilePath | ReadCsvBuffer[bytes] | ReadCsvBuffer[str],
    tz: str | datetime.tzinfo | None = None,
    date_col: str = "dt",
    temp_col: str = "tempF",
    gzipped: bool = False,
    freq: str | None = None,
    **kwargs,
):
    """Load meter data from a CSV file and convert to a dataframe. Farenheit is assumed for building models.

    Note: This is an example of the default csv structure assumed.
        ```python
        dt,tempF
        2017-01-01T00:00:00+00:00,21
        2017-01-01T01:00:00+00:00,22.5
        2017-01-01T02:00:00+00:00,23.5
        ```

    Args:
        filepath_or_buffer: File path or object.
        tz: Timezone represented in the meter data. Ex: `UTC` or `US/Pacific`
        date_col: Date period start column.
        temp_col: Temperature column.
        gzipped: Whether file is gzipped.
        freq: If given, apply frequency to data using `pandas.DataFrame.resample`. One of `['hourly', 'daily']`.
        **kwargs: Extra keyword arguments to pass to `pandas.read_csv`, such as `sep='|'`.
    """
    read_csv_kwargs = {
        "usecols": [date_col, temp_col],
        "dtype": {temp_col: np.float64},
        "parse_dates": [date_col],
        "index_col": date_col,
    }

    if gzipped:
        read_csv_kwargs.update({"compression": "gzip"})

    # allow passing extra kwargs
    read_csv_kwargs.update(kwargs)

    df = pd.read_csv(filepath_or_buffer, **read_csv_kwargs)
    df.index = pd.to_datetime(df.index, utc=True)

    # for pandas<0.24, which doesn't localize even with utc=True
    if df.index.tz is None:
        df.index = df.index.tz_localize("UTC")  # pragma: no cover

    if tz is not None:
        df = df.tz_convert(tz)

    if freq == "hourly":
        df = df.resample("h").sum(min_count=1)

    return df[temp_col]


def meter_data_from_json(data: list, orient: str = "list") -> pd.DataFrame:
    """Load meter data from a list of dictionary objects or a list of lists.

    Args:
        data: A list of meter data, with each row representing a single record.
        orient: Format of `data` parameter. Must be one of `['list', 'records']`.
            `'list'` is a list of lists, with the first element as start date and the second element as meter usage. `'records'` is a list of dicts.

    Note: This is an example of the default `list` structure.
        ```python
        [
            ['2017-01-01T00:00:00+00:00', 3.5],
            ['2017-02-01T00:00:00+00:00', 0.4],
            ['2017-03-01T00:00:00+00:00', 0.46],
        ]
        ```

    Note: This is an example of the `records` structure.
        ```python
        [
            {'start': '2017-01-01T00:00:00+00:00', 'value': 3.5},
            {'start': '2017-02-01T00:00:00+00:00', 'value': 0.4},
            {'start': '2017-03-01T00:00:00+00:00', 'value': 0.46},
        ]
        ```

    Returns:
        DataFrame with a single column (``'value'``) and a `pandas.DatetimeIndex`. A second column (``'estimated'``) may also be included if the input data contained an estimated boolean flag.
    """

    def _empty_meter_data_dataframe():
        return pd.DataFrame(
            {"value": []}, index=pd.DatetimeIndex([], tz="UTC", name="start")
        )

    if data is None:
        return _empty_meter_data_dataframe()

    if orient == "list":
        df = pd.DataFrame(data, columns=["start", "value"])
        df["start"] = pd.to_datetime(df.start, utc=True)
        df = df.set_index("start")
        return df
    elif orient == "records":

        def _noneify_meter_data_row(row):
            value = row["value"]
            if value is not None:
                try:
                    value = float(value)
                except ValueError:
                    value = None
            out_row = {"start": row["start"], "value": value}
            if "estimated" in row:
                estimated = row.get("estimated")
                out_row["estimated"] = estimated in [True, "true", "True", 1, "1"]
            return out_row

        noneified_data = [_noneify_meter_data_row(row) for row in data]
        df = pd.DataFrame(noneified_data)
        if df.empty:
            return _empty_meter_data_dataframe()
        df["start"] = pd.to_datetime(df.start, utc=True)
        df = df.set_index("start")
        df["value"] = df["value"].astype(float)
        if "estimated" in df.columns:
            df["estimated"] = (
                df["estimated"].where(df["estimated"].notna(), False).astype(bool)
            )
        return df
    else:
        raise ValueError("orientation not recognized.")


def temperature_data_from_json(data: list, orient: str = "list") -> pd.Series:
    """Load temperature data from json to a Series. Farenheit is assumed for building models.

    Args:
        data: A list of temperature data, with each row representing a single record.
        orient: Format of `data` parameter. Must be `'list'`.
            `'list'` is a list of lists, with the first element as start date and the second element as temperature.

    Note: This is an example of the default `list` structure.
        ```python
        [
            ['2017-01-01T00:00:00+00:00', 3.5],
            ['2017-01-01T01:00:00+00:00', 5.4],
            ['2017-01-01T02:00:00+00:00', 7.4],
        ]
        ```

    Returns:
        DataFrame with a single column (``'tempF'``) and a `pandas.DatetimeIndex`.

    Raises:
        ValueError: If `orient` is not `'list'`.
    """
    if orient == "list":
        df = pd.DataFrame(data, columns=["dt", "tempF"])
        series = df.tempF
        series.index = pd.to_datetime(df.dt, utc=True)
        return series
    else:
        raise ValueError("orientation not recognized.")


def meter_data_to_csv(
    meter_data: pd.DataFrame | pd.Series,
    path_or_buf: str | FilePath | WriteBuffer[bytes] | WriteBuffer[str],
) -> None:
    """Write meter data from a DataFrame or Series to a CSV. See also `pandas.DataFrame.to_csv`.

    Args:
        meter_data: DataFrame or Series with a ``'value'`` column and a `pandas.DatetimeIndex`.
        path_or_buf: Path or file handle.
    """
    if meter_data.index.name is None:
        meter_data.index.name = "start"

    return meter_data.to_csv(path_or_buf, index=True)


def temperature_data_to_csv(
    temperature_data: pd.Series,
    path_or_buf: str | FilePath | WriteBuffer[bytes] | WriteBuffer[str],
) -> None:
    """Write temperature data to CSV. See also :any:`pandas.DataFrame.to_csv`.

    Args:
        temperature_data: Temperature data series with :any:`pandas.DatetimeIndex`.
        path_or_buf: Path or file handle.
    """
    if temperature_data.index.name is None:
        temperature_data.index.name = "dt"
    if temperature_data.name is None:
        temperature_data.name = "temperature"

    return temperature_data.to_frame().to_csv(path_or_buf, index=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from io import StringIO
import pandas as pd
import statsmodels.formula.api as smf

from opendsm.eemeter.common.features import (
    compute_occupancy_feature,
    compute_temperature_bin_features,
    compute_time_features,
    merge_features,
)
from opendsm.eemeter.common.warnings import EEMeterWarning
from opendsm.eemeter.models.hourly_caltrack.metrics import ModelMetrics
from opendsm.eemeter.models.hourly_caltrack.segmentation import (
    CalTRACKSegmentModel,
    SegmentedModel,
    fit_model_segments,
)

__all__ = (
    "CalTRACKHourlyModelResults",
    "CalTRACKHourlyModel",
    "caltrack_hourly_fit_feature_processor",
    "caltrack_hourly_prediction_feature_processor",
    "fit_caltrack_hourly_model_segment",
    "fit_caltrack_hourly_model",
)


class CalTRACKHourlyModelResults(object):
    """Contains information about the chosen model.

    Attributes
    ----------
    status : :any:`str`
        A string indicating the status of this result. Possible statuses:

        - ``'NO DATA'``: No baseline data was available.
        - ``'NO MODEL'``: A complete model could not be constructed.
        - ``'SUCCESS'``: A model was constructed.
    method_name : :any:`str`
        The name of the method used to fit the baseline model.
    model : :any:`eemeter.CalTRACKHourlyModel` or :any:`None`
        The selected model, if any.
    warnings : :any:`list` of :any:`eemeter.EEMeterWarning`
        A list of any warnings reported during the model selection and fitting
        process.
    metadata : :any:`dict`
        An arbitrary dictionary of metadata to be associated with this result.
        This can be used, for example, to tag the results with attributes like
        an ID::

            {
                'id': 'METER_12345678',
            }

    settings : :any:`dict`
        A dictionary of settings used by the method.
    totals_metrics : :any:`ModelMetrics`
        A ModelMetrics object, if one is calculated and associated with this
        model. (This initializes to None.) The ModelMetrics object contains
        model fit information and descriptive statistics about the underlying data,
        with that data expressed as period totals.
    avgs_metrics : :any:`ModelMetrics`
        A ModelMetrics object, if one is calculated and associated with this
        model. (This initializes to None.) The ModelMetrics object contains
        model fit information and descriptive statistics about the underlying data,
        with that data expressed as daily averages.
    """

    def __init__(
        self, status, method_name, model=None, warnings=[], metadata=None, settings=None
    ):
        self.status = status
        self.method_name = method_name

        self.model = model

        self.warnings = warnings

        if metadata is None:
            metadata = {}
        self.metadata = metadata

        if settings is None:
            settings = {}
        self.settings = settings

        self.totals_metrics = None
        self.avgs_metrics = None

    def __repr__(self):
        return "CalTRACKHourlyModelResults(status='{}', method_name='{}')".format(
            self.status, self.method_name
        )

    def json(self, with_candidates=False):
        """Return a JSON-serializable representation of this result.

        The output of this function can be converted to a serialized string
        with :any:`json.dumps`.
        """

        def _json_or_none(obj):
            return None if obj is None else obj.json()

        def _json_or_none_in_dict(obj):
            return (
                None
                if obj is None
                else {key: _json_or_none(val) for key, val in obj.items()}
            )

        data = {
            "status": self.status,
            "method_name": self.method_name,
            "model": _json_or_none(self.model),
            "warnings": [w.json() for w in self.warnings],
            "metadata": self.metadata,
            "settings": self.settings,
            "totals_metrics": _json_or_none_in_dict(self.totals_metrics),
            "avgs_metrics": _json_or_none_in_dict(self.avgs_metrics),
        }
        return data

    @classmethod
    def from_json(cls, data):
        """Loads a JSON-serializable representation into the model state.

        The input of this function is a dict which can be the result
        of :any:`json.loads`.
        """

        # "model" is a CalTRACKHourlyModel that was serialized
        model = None
        d = data.get("model")
        if d:
            model = CalTRACKHourlyModel.from_json(d)

        c = cls(
            data.get("status"),
            data.get("method_name"),
            model=model,
            warnings=data.get("warnings"),
            metadata=data.get("metadata"),
            settings=data.get("settings"),
        )

        # Note the metrics do not contain all the data needed
        # for reconstruction (like the input pandas) ...
        d = data.get("avgs_metrics")
        if d:
            c.avgs_metrics = ModelMetrics.from_json(d)  # pragma: no cover
            c.avgs_metrics = {
                segment_name: ModelMetrics.from_json(seg_d)
                for segment_name, seg_d in d.items()
            }
        d = data.get("totals_metrics")
        if d:
            c.totals_metrics = ModelMetrics.from_json(d)  # pragma: no cover
            c.totals_metrics = {
                segment_name: ModelMetrics.from_json(seg_d)
                for segment_name, seg_d in d.items()
            }
        return c

    def predict(self, prediction_index, temperature_data, **kwargs):
        """Predict over a particular index using temperature data.

        Parameters
        ----------
        prediction_index : :any:`pandas.DatetimeIndex`
            Time period over which to predict.
        temperature_data : :any:`pandas.DataFrame`
            Hourly temperature data to use for prediction. Time period should match
            the ``prediction_index`` argument.
        **kwargs
            Extra keyword arguments to send to self.model.predict

        Returns
        -------
        prediction : :any:`pandas.DataFrame`
            The predicted usage values.
        """
        return self.model.predict(prediction_index, temperature_data, **kwargs)


class _PredictionSegmentInfo:
    """
    Class to handle the different segment_type parameters
    that provides the correct values to the CalTrackHourlyModel initialization.
    """

    def __init__(self, segment_type: str):
        if segment_type not in ["single", "three_month_weighted"]:
            raise ValueError("segment type must be single or three_month_weighted")

        if segment_type == "single":
            self.prediction_segment_type = segment_type
            self.prediction_segment_name_mapping = None
            return

        if segment_type == "three_month_weighted":
            self.prediction_segment_name_mapping = {
                "jan": "dec-jan-feb-weighted",
                "feb": "jan-feb-mar-weighted",
                "mar": "feb-mar-apr-weighted",
                "apr": "mar-apr-may-weighted",
                "may": "apr-may-jun-weighted",
                "jun": "may-jun-jul-weighted",
                "jul": "jun-jul-aug-weighted",
                "aug": "jul-aug-sep-weighted",
                "sep": "aug-sep-oct-weighted",
                "oct": "sep-oct-nov-weighted",
                "nov": "oct-nov-dec-weighted",
                "dec": "nov-dec-jan-weighted",
            }
            self.prediction_segment_type = "one_month"
            return


class CalTRACKHourlyModel(SegmentedModel):
    """An object which holds CalTRACK Hourly model data and metadata, and
    which can be used for prediction.

    Attributes
    ----------
    segment_models : :any:`dict` of `eemeter.CalTRACKSegmentModel`
        Dictionary of models for each segment, keys are segment names.
    occupancy_lookup : :any:`pandas.DataFrame`
        A dataframe with occupancy flags for each hour of the week and each segment.
        Segment names are columns, occupancy flags are 0 or 1.
    occupied_temperature_bins : :any:`pandas.DataFrame`
        A dataframe of bin endpoint flags for each segment. Segment names are columns.
    unoccupied_temperature_bins : :any:`pandas.DataFrame`
        Ditto for the unoccupied mode.
    segment_type : :any:`str`
        The type of segment used to fit the model
    """

    def __init__(
        self,
        segment_models,
        occupancy_lookup,
        occupied_temperature_bins,
        unoccupied_temperature_bins,
        segment_type: str,
    ):
        self.occupancy_lookup = occupancy_lookup
        self.occupied_temperature_bins = occupied_temperature_bins
        self.unoccupied_temperature_bins = unoccupied_temperature_bins
        self.segment_type = segment_type

        prediction_info = _PredictionSegmentInfo(segment_type=segment_type)
        super(CalTRACKHourlyModel, self).__init__(
            segment_models=segment_models,
            prediction_segment_type=prediction_info.prediction_segment_type,
            prediction_segment_name_mapping=prediction_info.prediction_segment_name_mapping,
            prediction_feature_processor=caltrack_hourly_prediction_feature_processor,
            prediction_feature_processor_kwargs={
                "occupancy_lookup": self.occupancy_lookup,
                "occupied_temperature_bins": self.occupied_temperature_bins,
                "unoccupied_temperature_bins": self.unoccupied_temperature_bins,
            },
        )

    def json(self):
        """Return a JSON-serializable representation of this result.

        The output of this function can be converted to a serialized string
        with :any:`json.dumps`.
        """
        data = super(CalTRACKHourlyModel, self).json()
        data.update(
            {
                "occupancy_lookup": self.occupancy_lookup.to_json(orient="split"),
                "occupied_temperature_bins": self.occupied_temperature_bins.to_json(
                    orient="split"
                ),
                "unoccupied_temperature_bins": self.unoccupied_temperature_bins.to_json(
                    orient="split"
                ),
                "segment_type": self.segment_type,
            }
        )
        return data

    @classmethod
    def from_json(cls, data):
        """Loads a JSON-serializable representation into the model state.

        The input of this function is a dict which can be the result
        of :any:`json.loads`.
        """

        segment_models = [
            CalTRACKSegmentModel.from_json(s) for s in data.get("segment_models")
        ]

        occupancy_lookup = pd.read_json(
            StringIO(data.get("occupancy_lookup")), orient="split"
        )
        occupancy_lookup.index = occupancy_lookup.index.astype("category")

        c = cls(
            segment_models,
            occupancy_lookup,
            pd.read_json(
                StringIO(data.get("occupied_temperature_bins")), orient="split"
            ),
            pd.read_json(
                StringIO(data.get("unoccupied_temperature_bins")), orient="split"
            ),
            data.get("segment_type"),
        )

        return c


def caltrack_hourly_fit_feature_processor(
    segment_name,
    segmented_data,
    occupancy_lookup,
    occupied_temperature_bins,
    unoccupied_temperature_bins,
):
    """A function that takes in temperature data and returns a dataframe of
    features suitable for use with :any:`eemeter.fit_caltrack_hourly_model_segment`.
    Designed for use with :any:`eemeter.iterate_segmented_dataset`.

    Parameters
    ----------
    segment_name : :any:`str`
        The name of the segment.
    segmented_data : :any:`pandas.DataFrame`
        Hourly temperature data for the segment.
    occupancy_lookup : :any:`pandas.DataFrame`
        A dataframe with occupancy flags for each hour of the week and each segment.
        Segment names are columns, occupancy flags are 0 or 1.
    occupied_temperature_bins : :any:`pandas.DataFrame`
        A dataframe of bin endpoint flags for each segment. Segment names are columns.
    unoccupied_temperature_bins : :any:`pandas.DataFrame`
        Ditto for the unoccupied mode.

    Returns
    -------
    features : :any:`pandas.DataFrame`
        A dataframe of features with the following columns:

        - 'meter_value': the observed meter value
        - 'hour_of_week': 0-167
        - 'bin_<0-6>_occupied': temp bin feature, or 0 if unoccupied
        - 'bin_<0-6>_unoccupied': temp bin feature or 0 in occupied
        - 'weight': 0.0 or 0.5 or 1.0
    """
    # get occupied feature
    hour_of_week = segmented_data.hour_of_week
    occupancy = occupancy_lookup[segment_name]
    occupancy_feature = compute_occupancy_feature(hour_of_week, occupancy)

    # get temperature bin features
    temperatures = segmented_data.temperature_mean
    occupied_bin_endpoints_list = (
        occupied_temperature_bins[segment_name]
        .index[occupied_temperature_bins[segment_name]]
        .tolist()
    )
    unoccupied_bin_endpoints_list = (
        unoccupied_temperature_bins[segment_name]
        .index[unoccupied_temperature_bins[segment_name]]
        .tolist()
    )
    occupied_temperature_bin_features = compute_temperature_bin_features(
        segmented_data.temperature_mean, occupied_bin_endpoints_list
    )
    occupied_temperature_bin_features[occupancy_feature == 0] = 0
    occupied_temperature_bin_features.rename(
        columns={
            c: "{}_occupied".format(c)
            for c in occupied_temperature_bin_features.columns
        },
        inplace=True,
    )
    unoccupied_temperature_bin_features = compute_temperature_bin_features(
        segmented_data.temperature_mean, unoccupied_bin_endpoints_list
    )
    unoccupied_temperature_bin_features[occupancy_feature == 1] = 0
    unoccupied_temperature_bin_features.rename(
        columns={
            c: "{}_unoccupied".format(c)
            for c in unoccupied_temperature_bin_features.columns
        },
        inplace=True,
    )

    # combine features
    return merge_features(
        [
            segmented_data[["meter_value", "hour_of_week"]],
            occupied_temperature_bin_features,
            unoccupied_temperature_bin_features,
            segmented_data.weight,
        ]
    )


def caltrack_hourly_prediction_feature_processor(
    segment_name,
    segmented_data,
    occupancy_lookup,
    occupied_temperature_bins,
    unoccupied_temperature_bins,
):
    """A function that takes in temperature data and returns a dataframe of
    features suitable for use inside :any:`eemeter.CalTRACKHourlyModel`.
    Designed for use with :any:`eemeter.iterate_segmented_dataset`.

    Parameters
    ----------
    segment_name : :any:`str`
        The name of the segment.
    segmented_data : :any:`pandas.DataFrame`
        Hourly temperature data for the segment.
    occupancy_lookup : :any:`pandas.DataFrame`
        A dataframe with occupancy flags for each hour of the week and each segment.
        Segment names are columns, occupancy flags are 0 or 1.
    occupied_temperature_bins : :any:`pandas.DataFrame`
        A dataframe of bin endpoint flags for each segment. Segment names are columns.
    unoccupied_temperature_bins : :any:`pandas.DataFrame`
        Ditto for the unoccupied mode.

    Returns
    -------
    features : :any:`pandas.DataFrame`
        A dataframe of features with the following columns:

        - 'hour_of_week': 0-167
        - 'bin_<0-6>_occupied': temp bin feature, or 0 if unoccupied
        - 'bin_<0-6>_unoccupied': temp bin feature or 0 in occupied
        - 'weight': 1
    """
    # hour of week feature
    hour_of_week_feature = compute_time_features(
        segmented_data.index, hour_of_week=True, day_of_week=False, hour_of_day=False
    )

    # occupancy feature
    occupancy = occupancy_lookup[segment_name]
    occupancy_feature = compute_occupancy_feature(
        hour_of_week_feature.hour_of_week, occupancy
    )

    # get temperature bin features
    temperatures = segmented_data
    occupied_bin_endpoints_list = (
        occupied_temperature_bins[segment_name]
        .index[occupied_temperature_bins[segment_name]]
        .tolist()
    )
    unoccupied_bin_endpoints_list = (
        unoccupied_temperature_bins[segment_name]
        .index[unoccupied_temperature_bins[segment_name]]
        .tolist()
    )
    occupied_temperature_bin_features = compute_temperature_bin_features(
        segmented_data.temperature_mean, occupied_bin_endpoints_list
    )
    occupied_temperature_bin_features[occupancy_feature == 0] = 0
    occupied_temperature_bin_features.rename(
        columns={
            c: "{}_occupied".format(c)
            for c in occupied_temperature_bin_features.columns
        },
        inplace=True,
    )
    unoccupied_temperature_bin_features = compute_temperature_bin_features(
        segmented_data.temperature_mean, unoccupied_bin_endpoints_list
    )
    unoccupied_temperature_bin_features[occupancy_feature == 1] = 0
    unoccupied_temperature_bin_features.rename(
        columns={
            c: "{}_unoccupied".format(c)
            for c in unoccupied_temperature_bin_features.columns
        },
        inplace=True,
    )

    # combine features
    return merge_features(
        [
            hour_of_week_feature,
            occupied_temperature_bin_features,
            unoccupied_temperature_bin_features,
            segmented_data.weight,
        ]
    )


def fit_caltrack_hourly_model_segment(segment_name, segment_data):
    """Fit a model for a single segment.

    Parameters
    ----------
    segment_name : :any:`str`
        The name of the segment.
    segment_data : :any:`pandas.DataFrame`
        A design matrix for caltrack hourly, of the form returned by
        :any:`eemeter.caltrack_hourly_prediction_feature_processor`.

    Returns
    -------
    segment_model : :any:`CalTRACKSegmentModel`
        A model that represents the fitted model.
    """

    warnings = []
    if segment_data.dropna().empty:
        model = None
        formula = None
        model_params = None
        warnings.append(
            EEMeterWarning(
                qualified_name="eemeter.fit_caltrack_hourly_model_segment.no_nonnull_data",
                description="The segment contains either an empty dataset or all NaNs.",
                data={
                    "n_rows": segment_data.shape[0],
                    "n_rows_after_dropna": segment_data.dropna().shape[0],
                },
            )
        )

    else:

        def _get_hourly_model_formula(data):
            return "meter_value ~ C(hour_of_week) - 1{}".format(
                "".join(
                    [" + {}".format(c) for c in data.columns if c.startswith("bin")]
                )
            )

        formula = _get_hourly_model_formula(segment_data)

        # remove categories that only have null or missing entries
        # this ensures that predictions will predict null
        segment_data["hour_of_week"] = pd.Categorical(
            segment_data["hour_of_week"],
            categories=segment_data["hour_of_week"].dropna().unique(),
            ordered=False,
        )
        model = smf.wls(formula=formula, data=segment_data, weights=segment_data.weight)
        model_params = {coeff: value for coeff, value in model.fit().params.items()}

    segment_model = CalTRACKSegmentModel(
        segment_name=segment_name,
        model=model,
        formula=formula,
        model_params=model_params,
        warnings=warnings,
    )
    if model:
        this_segment_data = segment_data[segment_data.weight == 1]
        predicted_value = pd.Series(model.fit().predict(this_segment_data))
        segment_model.totals_metrics = ModelMetrics(
            this_segment_data.meter_value, predicted_value, len(model_params)
        )
    else:
        segment_model.totals_metrics = None

    return segment_model


def fit_caltrack_hourly_model(
    segmented_design_matrices,
    occupancy_lookup,
    occupied_temperature_bins,
    unoccupied_temperature_bins,
    segment_type: str,
):
    """Fit a CalTRACK hourly model

    Parameters
    ----------
    segmented_design_matrices : :any:`dict` of :any:`pandas.DataFrame`
        A dictionary of dataframes of the form returned by
        :any:`eemeter.create_caltrack_hourly_segmented_design_matrices`
    occupancy_lookup : :any:`pandas.DataFrame`
        A dataframe with occupancy flags for each hour of the week and each segment.
        Segment names are columns, occupancy flags are 0 or 1.
    occupied_temperature_bins : :any:`pandas.DataFrame`
        A dataframe of bin endpoint flags for each segment. Segment names are columns.
    unoccupied_temperature_bins : :any:`pandas.DataFrame`
        Ditto for the unoccupied mode.

    Returns
    -------
    model : :any:`CalTRACKHourlyModelResults`
        Has a `model.predict` method which take input data and makes a prediction
        using this model.
    """
    segment_models = fit_model_segments(
        segmented_design_matrices, fit_caltrack_hourly_model_segment
    )
    all_warnings = [
        warning
        for segment_model in segment_models
        for warning in segment_model.warnings
    ]

    model = CalTRACKHourlyModel(
        segment_models,
        occupancy_lookup,
        occupied_temperature_bins,
        unoccupied_temperature_bins,
        segment_type,
    )

    model_results = CalTRACKHourlyModelResults(
        status="SUCCEEDED",
        method_name="caltrack_hourly",
        warnings=all_warnings,
        model=model,
    )
    model_results.totals_metrics = {
        seg_model.segment_name: seg_model.totals_metrics for seg_model in segment_models
    }
    return model_results



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import json

import pytz
from dateutil.parser import parse as parse_date
import importlib.resources

from ..utilities.io import meter_data_from_csv, temperature_data_from_csv

__all__ = ("samples", "load_sample")


def _load_sample_metadata():
    with importlib.resources.files("opendsm.eemeter.samples").joinpath(
        "metadata.json"
    ).open("rb") as f:
        metadata = json.loads(f.read())
    return metadata


def samples():
    """Load a list of sample data identifiers.

    Returns
    -------
    samples : :any:`list` of :any:`str`
        List of sample identifiers for use with :any:`opendsm.eemeter.load_sample`.
    """
    sample_metadata = _load_sample_metadata()
    return list(sorted(sample_metadata.keys()))


def load_sample(sample, tempF=True):
    """Load meter data, temperature data, and metadata for associated with a
    particular sample identifier. Note: samples are simulated, not real, data.

    Parameters
    ----------
    sample : :any:`str`
        Identifier of sample. Complete list can be obtained with
        :any:`opendsm.eemeter.samples`.
    tempF : :any 'bool'
        Flag regarding whether the sample temperature dataset is associated with Fahrenheit or Celsius.

    Returns
    -------
    meter_data, temperature_data, metadata : :any:`tuple` of :any:`pandas.DataFrame`, :any:`pandas.Series`, and :any:`dict`
        Meter data, temperature data, and metadata for this sample identifier.
    """
    if tempF == True:
        temp_units = "tempF"
    else:
        temp_units = "tempC"

    sample_metadata = _load_sample_metadata()
    metadata = sample_metadata.get(sample)
    if metadata is None:
        raise ValueError(
            "Sample not found: {}. Try one of these?\n{}".format(
                sample,
                "\n".join(
                    [" - {}".format(key) for key in sorted(sample_metadata.keys())]
                ),
            )
        )

    freq = metadata.get("freq")
    if freq not in ("hourly", "daily"):
        freq = None

    meter_data_filename = metadata["meter_data_filename"]
    with importlib.resources.files("opendsm.eemeter.samples").joinpath(
        meter_data_filename
    ).open("rb") as f:
        meter_data = meter_data_from_csv(f, gzipped=True, freq=freq)

    temperature_filename = metadata["temperature_filename"]
    with importlib.resources.files("opendsm.eemeter.samples").joinpath(
        temperature_filename
    ).open("rb") as f:
        temperature_data = temperature_data_from_csv(
            f, gzipped=True, freq="hourly", temp_col=temp_units
        )

    metadata["blackout_start_date"] = pytz.UTC.localize(
        parse_date(metadata["blackout_start_date"])
    )
    metadata["blackout_end_date"] = pytz.UTC.localize(
        parse_date(metadata["blackout_end_date"])
    )

    return meter_data, temperature_data, metadata



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from typing import Optional

import numba
import numpy as np

from opendsm.eemeter.models.daily.base_models.full_model import (
    full_model,
    full_model_weight,
)
from opendsm.eemeter.models.daily.objective_function import obj_fcn_decorator
from opendsm.eemeter.models.daily.optimize import Optimizer
from opendsm.eemeter.models.daily.parameters import ModelCoefficients, ModelType
from opendsm.eemeter.models.daily.utilities.base_model import fix_identical_bnds


def fit_tidd(
    T,
    obs,
    weights,
    settings,
    opt_options,
    x0: Optional[ModelCoefficients] = None,
    bnds=None,
    initial_fit=False,
):
    if x0 is None:
        x0 = _tidd_x0(T, obs)

    if initial_fit:
        alpha = settings.alpha_selection
    else:
        alpha = settings.alpha_final

    intercept_bnds = np.quantile(obs, [0.01, 0.99])
    bnds_0 = np.array([intercept_bnds])

    if bnds is None:
        bnds = bnds_0

    bnds = _tidd_update_bnds(bnds, bnds_0)

    coef_id = ["intercept"]
    model_fcn = _tidd
    weight_fcn = _tidd_weight
    TSS_fcn = _tidd_total_sum_of_squares
    obj_fcn = obj_fcn_decorator(
        model_fcn, weight_fcn, TSS_fcn, T, obs, weights, settings, alpha, coef_id, initial_fit
    )

    res = Optimizer(
        obj_fcn, x0.to_np_array(), bnds, coef_id, settings, opt_options
    ).run()

    return res


# Model Functions
def _tidd_x0(T, obs):
    intercept = np.median(obs)
    return ModelCoefficients(model_type=ModelType.TIDD, intercept=intercept)


@numba.jit(nopython=True, error_model="numpy", cache=True)
def set_full_model_coeffs(intercept):
    hdd_bp = hdd_beta = hdd_k = cdd_bp = cdd_beta = cdd_k = 0

    return np.array([hdd_bp, hdd_beta, hdd_k, cdd_bp, cdd_beta, cdd_k, intercept])


def _tidd(intercept, T_fit_bnds=np.array([]), T=np.array([])):
    model_vars = set_full_model_coeffs(intercept)

    return full_model(*model_vars, T_fit_bnds, T)


def _tidd_total_sum_of_squares(intercept, T, obs):
    TSS = np.sum((obs - np.mean(obs)) ** 2)

    return TSS


def _tidd_update_bnds(new_bnds, bnds):
    new_bnds = bnds

    new_bnds = np.sort(new_bnds, axis=1)
    new_bnds = fix_identical_bnds(new_bnds)

    return new_bnds


def _tidd_weight(
    intercept, T, residual, sigma=3.0, quantile=0.25, alpha=2.0, min_weight=0.0
):
    model_vars = set_full_model_coeffs(intercept)

    return full_model_weight(
        *model_vars, T, residual, sigma, quantile, alpha, min_weight
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import logging as _logging

from .__version__ import __title__, __description__, __url__, __version__
from .__version__ import __author__, __author_email__, __license__
from .__version__ import __copyright__

import platform
import warnings

# these happen during native code execution and segfault pytest when filterwarnings is set to error
warnings.filterwarnings("ignore", module="importlib._bootstrap")
warnings.filterwarnings(
    "ignore", "builtin type swigvarlink has no __module__ attribute"
)
warnings.filterwarnings(
    "ignore", "builtin type SwigPyPacked has no __module__ attribute"
)

if platform.system() == "Windows":
    # numba JIT breaks on Windows with int32/int64 return types
    from numba import config

    config.DISABLE_JIT = True

from .common import (
    abstract_data_processor,
    abstract_data_settings,
    adaptive_loss_Z,
    adaptive_loss,
    const,
    data_settings,
    test_data,
    utils,
)
from . import (
    eemeter,
    drmeter,
)

# Set default logging handler to avoid "No handler found" warnings.
_logging.getLogger(__name__).addHandler(_logging.NullHandler())

# exclude built-in imports from namespace
__all__ = (
    "__title__",
    "__description__",
    "__url__",
    "__version__",
    "__author__",
    "__author_email__",
    "__license__",
    "__copyright__",
    "abstract_data_processor",
    "abstract_data_settings",
    "adaptive_loss_Z",
    "adaptive_loss",
    "const",
    "data_settings",
    "test_data",
    "utils",
    "eemeter",
    "drmeter",
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from __future__ import annotations

import datetime

import numpy as np
import pandas as pd

import opendsm.common.const as _const
from opendsm.eemeter.common.data_processor_utilities import (
    as_freq,
    clean_billing_daily_data,
    compute_minimum_granularity,
    remove_duplicates,
)
from opendsm.eemeter.common.features import compute_temperature_features
from opendsm.eemeter.common.warnings import EEMeterWarning
from opendsm.eemeter.common.sufficiency_criteria import DailySufficiencyCriteria


class _DailyData:
    """Private base class for daily baseline and reporting data.

    Will raise exception during data sufficiency check if instantiated

    Args:
        df (pd.DataFrame): The DataFrame containing the observed meter data.
        is_electricity_data (bool): A flag indicating whether the data represents electricity data. This is required as electricity data with 0 values are converted to NaNs.
    """

    def __init__(self, df: pd.DataFrame, is_electricity_data: bool):
        self._df = None
        self.warnings = []
        self.disqualification = []
        self.is_electricity_data = is_electricity_data
        self.tz = None

        # TODO re-examine dq/warning pattern. keep consistent between
        # either implicitly setting as side effects, or returning and assigning outside
        self._df, temp_coverage = self._set_data(df)

        sufficiency_df = self._df.merge(
            temp_coverage, left_index=True, right_index=True, how="left"
        )
        disqualification, warnings = self._check_data_sufficiency(sufficiency_df)

        self.disqualification += disqualification
        self.warnings += warnings
        self.log_warnings()

    @property
    def df(self) -> pd.DataFrame | None:
        """Get the corrected input data stored in the class. The actual dataframe is immutable, this returns a copy."""

        if self._df is None:
            return None
        else:
            return self._df.copy()

    @classmethod
    def from_series(
        cls,
        meter_data: pd.Series | pd.DataFrame,
        temperature_data: pd.Series | pd.DataFrame,
        is_electricity_data: bool,
    ):
        """Create an instance of the Data class from meter data and temperature data.

        Public method that can can handle two separate series (meter and temperature) and join them to create a single dataframe. The temperature column should have values in Fahrenheit.

        Args:
            meter_data: The meter data.
            temperature_data: The temperature data.
            is_electricity_data: A flag indicating whether the data represents electricity data. This is required as electricity data with 0 values are converted to NaNs.

        Returns:
            An instance of the Data class with the dataframe populated with the corrected data, along with warnings and disqualifications based on the input.
        """
        if isinstance(meter_data, pd.Series):
            meter_data = meter_data.to_frame()
        if isinstance(temperature_data, pd.Series):
            temperature_data = temperature_data.to_frame()
        meter_data = meter_data.rename(columns={meter_data.columns[0]: "observed"})
        temperature_data = temperature_data.rename(
            columns={temperature_data.columns[0]: "temperature"}
        )
        temperature_data.index = temperature_data.index.tz_convert(
            meter_data.index.tzinfo
        )

        if temperature_data.empty:
            raise ValueError("Temperature data cannot be empty.")
        if meter_data.empty:
            # reporting from_series always passes a full index of nan
            raise ValueError("Meter data cannot by empty.")

        is_billing_data = False
        if not meter_data.empty:
            is_billing_data = compute_minimum_granularity(
                meter_data.index, "billing"
            ).startswith("billing")

        # first, trim the data to exclude NaNs on the outer edges of the data
        last_meter_index = meter_data.last_valid_index()
        if is_billing_data:
            # preserve final NaN for billing data only
            last = meter_data.last_valid_index()
            if last and last != meter_data.index[-1]:
                # TODO include warning here for non-NaN final billing row since it will be discarded
                last_meter_index = meter_data.index[meter_data.index.get_loc(last) + 1]
        meter_data = meter_data.loc[meter_data.first_valid_index() : last_meter_index]
        temperature_data = temperature_data.loc[
            temperature_data.first_valid_index() : temperature_data.last_valid_index()
        ]

        # TODO consider a refactor of the period offset calculation/slicing.
        # it seems like a fairly dense block of code for something conceptually simple.
        # at the very least, try to clarify variable names a bit

        period_diff_first = pd.Timedelta(0)
        period_diff_last = pd.Timedelta(0)
        # calculate difference in period length for first and last rows in meter/temp
        # first/last will generally be the same offset for daily/hourly, but billing can be quite variable
        # could consider using to_offset(index.inferred_freq) if available,
        # but the intent here is just to provide a lenient first trim.
        # checking for consistent frequency is done later during __init__
        if len(meter_data.index) > 1 and len(temperature_data.index) > 1:
            period_meter_first = meter_data.index[1] - meter_data.index[0]
            period_temp_first = temperature_data.index[1] - temperature_data.index[0]
            period_diff_first = period_meter_first - period_temp_first

            period_meter_last = meter_data.index[-1] - meter_data.index[-2]
            period_temp_last = temperature_data.index[-1] - temperature_data.index[-2]
            period_diff_last = period_meter_last - period_temp_last

        # if diff is positive, meter period is longer (lower frequency)
        zero_offset = pd.Timedelta(0)
        meter_period_first_longer = period_diff_first > zero_offset
        meter_period_last_longer = period_diff_last > zero_offset

        # large period needs a buffer for the min index, and no buffer for the max index
        # short period needs a buffer for the max index, and no buffer for the min index
        meter_offset_first = (
            period_diff_first if meter_period_first_longer else zero_offset
        )
        meter_offset_last = (
            -period_diff_last if not meter_period_last_longer else zero_offset
        )
        temp_offset_first = (
            -period_diff_first if not meter_period_first_longer else zero_offset
        )
        temp_offset_last = period_diff_last if meter_period_last_longer else zero_offset

        # if the shorter period ends on an exact index of the longer, we accept it.
        # the data should be DQ'd later due to insufficiency for the period

        # constrain meter index to temperature index
        temp_index_min = temperature_data.index.min() - meter_offset_first
        temp_index_max = temperature_data.index.max() + meter_offset_last
        meter_data = meter_data[temp_index_min:temp_index_max]
        if meter_data.empty:
            raise ValueError("Meter and temperature data are fully misaligned.")

        # if billing detected, subtract one day from final index since dataframe input assumes final row is part of period
        if is_billing_data:
            new_index = meter_data.index[:-1].union(
                [(meter_data.index[-1] - pd.Timedelta(days=1))]
            )
            if len(new_index) == len(meter_data.index):
                meter_data.index = new_index
            else:
                # handles the case of a 1 day off-cycle read at end of series
                meter_data = meter_data[:-1]

        # constrain temperature index to meter index
        meter_index_min = meter_data.index.min() - temp_offset_first
        meter_index_max = meter_data.index.max() + temp_offset_last
        if is_billing_data and len(meter_data) > 1:
            # last billing period is offset by one index
            meter_index_max = meter_data.index[-2] + temp_offset_last
        temperature_data = temperature_data[meter_index_min:meter_index_max]

        if is_billing_data:
            # TODO consider adding misaligned data warning here if final row was not already NaN
            meter_data.iloc[-1] = np.nan

        df = pd.concat([meter_data, temperature_data], axis=1)
        return cls(df, is_electricity_data)

    def log_warnings(self) -> None:
        """Logs the warnings and disqualifications associated with the data.

        View the disqualifications and warnings associated with the current data input provided.

        Returns:
            None
        """
        for warning in self.warnings + self.disqualification:
            warning.warn()

    def _compute_meter_value_df(self, df: pd.DataFrame):
        """
        Computes the meter value DataFrame by cleaning and processing the observed meter data.
        1. The minimum granularity is computed from the non null rows.
        2. The meter data is cleaned and downsampled/upsampled into the correct frequency using clean_billing_daily_data()
        3. Add missing days as NaN by merging with a full year daily index.

        Parameters
        ----------

            df (pd.DataFrame): The DataFrame containing the observed meter data.

        Returns
        -------
            pd.DataFrame: The cleaned and processed meter value DataFrame.
        """
        meter_series = df["observed"].dropna()
        if meter_series.empty:
            return df["observed"].resample("D").first().to_frame()

        # Dropping the NaNs is beneficial when the meter data is spread over hourly temperature data, causing lots of NaNs
        # But causes problems in detection of frequency when there are genuine missing values. The missing days are accounted for in the sufficiency_criteria_baseline method
        # whereas they should actually be kept.
        start_date = df.index.min()
        end_date = df.index.max()
        min_granularity = compute_minimum_granularity(meter_series.index, "daily")
        if min_granularity.startswith("billing"):
            # TODO : make this a warning instead of an exception
            raise ValueError("Billing data is not allowed in the daily model")
        meter_value_df = clean_billing_daily_data(
            meter_series, min_granularity, self.warnings
        )

        meter_value_df = meter_value_df.rename(columns={"value": "observed"})

        # To account for the above issue, we create an index with all the days and then merge the meter_value_df with it
        # This will ensure that the missing days are kept in the dataframe
        # Create an index with all the days from the start and end date of 'meter_value_df'
        all_days_index = pd.date_range(
            start=start_date,
            end=end_date,
            freq="D",
            tz=df.index.tz,
            ambiguous=True,
            nonexistent="shift_forward",
        )
        all_days_df = pd.DataFrame(index=all_days_index)
        # the following drops common days to handle DST issues with pytz.
        # doesn't seem to be a problem with ZoneInfo, so we can
        # probably handle this better once 3.8 is EOL and we disallow pytz tzinfo.
        # TODO regardless, it feels like there should be a better way to match
        # the indices on date than by comparing strftime in this manner
        all_days_df = all_days_df[
            ~all_days_df.index.strftime("%Y%m%d").isin(
                meter_series.index.strftime("%Y%m%d")
            )
        ]
        meter_value_df = meter_value_df.merge(
            all_days_df, left_index=True, right_index=True, how="outer"
        )

        return meter_value_df

    def _compute_temperature_features(
        self, df: pd.DataFrame, meter_index: pd.DatetimeIndex
    ):
        """
        Compute temperature features for the given DataFrame and meter index.
        1. The frequency of the temperature data is inferred and set to hourly if not already. If frequency is not inferred or its lower than hourly, a warning is added.
        2. The temperature data is downsampled/upsampled into the daily frequency using as_freq()
        3. High frequency temperature data is checked for missing values and a warning is added if more than 50% of the data is missing, and those rows are set to NaN.
        4. If frequency was already hourly, compute_temperature_features() is used to recompute the temperature to match with the meter index.

        Parameters
        ----------

            df (pd.DataFrame): The DataFrame containing temperature data.
            meter_index (pd.DatetimeIndex): The meter index.

        Returns
        -------

            pd.Series: The computed temperature values.
            pd.DataFrame: The computed temperature features.
        """
        temp_series = df["temperature"]
        temp_series.index.freq = temp_series.index.inferred_freq
        if temp_series.index.freq != "h":
            if temp_series.index.freq is None or temp_series.index.freq > pd.Timedelta(
                hours=1
            ):
                # Add warning for frequencies longer than 1 hour
                self.warnings.append(
                    EEMeterWarning(
                        qualified_name="eemeter.sufficiency_criteria.unable_to_confirm_daily_temperature_sufficiency",
                        description=(
                            "Cannot confirm that pre-aggregated temperature data had sufficient hours kept"
                        ),
                        data={},
                    )
                )
            if temp_series.index.freq != "D":
                # Downsample / Upsample the temperature data to daily
                temperature_features = as_freq(
                    temp_series, "D", series_type="instantaneous", include_coverage=True
                )
                # If high frequency data check for 50% data coverage in rollup
                if len(temperature_features[temperature_features.coverage <= 0.5]) > 0:
                    self.warnings.append(
                        EEMeterWarning(
                            qualified_name="eemeter.sufficiency_criteria.missing_high_frequency_temperature_data",
                            description=(
                                "More than 50% of the high frequency Temperature data is missing."
                            ),
                            data={
                                "high_frequency_data_missing_count": len(
                                    temperature_features[
                                        temperature_features.coverage <= 0.5
                                    ].index.to_list()
                                )
                            },
                        )
                    )

                # Set missing high frequency data to NaN
                temperature_features.loc[
                    temperature_features.coverage > 0.5, "value"
                ] = (
                    temperature_features[temperature_features.coverage > 0.5].value
                    / temperature_features[temperature_features.coverage > 0.5].coverage
                )

                temperature_features = (
                    temperature_features[temperature_features.coverage > 0.5]
                    .reindex(temperature_features.index)[["value"]]
                    .rename(columns={"value": "temperature_mean"})
                )

                if "coverage" in temperature_features.columns:
                    temperature_features = temperature_features.drop(
                        columns=["coverage"]
                    )
            else:
                temperature_features = temp_series.to_frame(name="temperature_mean")

            temperature_features["temperature_null"] = temp_series.isnull().astype(int)
            temperature_features["temperature_not_null"] = temp_series.notnull().astype(
                int
            )
            temperature_features["n_days_kept"] = 0  # unused
            temperature_features["n_days_dropped"] = 0  # unused
        else:
            # TODO hacky method of avoiding the last index nan convention
            if not meter_index.empty:
                buffer_idx = meter_index.max() + pd.Timedelta(days=1)
                meter_index = meter_index.union([buffer_idx])
            temperature_features = compute_temperature_features(
                meter_index,
                temp_series,
                data_quality=True,
            )
            temperature_features = temperature_features[:-1]

            # Only check for high frequency temperature data if it exists
            if (
                temperature_features.temperature_not_null
                + temperature_features.temperature_null
            ).median() > 1:
                invalid_temperature_rows = (
                    temperature_features.temperature_not_null
                    / (
                        temperature_features.temperature_not_null
                        + temperature_features.temperature_null
                    )
                ) <= 0.5

                # Set high frequency temperature data with more than 50% data missing as NaN
                if invalid_temperature_rows.any():
                    self.warnings.append(
                        EEMeterWarning(
                            qualified_name="eemeter.sufficiency_criteria.missing_high_frequency_temperature_data",
                            description=(
                                "More than 50% of the high frequency temperature data is missing."
                            ),
                            data=[
                                timestamp.isoformat()
                                for timestamp in invalid_temperature_rows[invalid_temperature_rows].index
                            ],
                        )
                    )
                    temperature_features.loc[
                        invalid_temperature_rows, "temperature_mean"
                    ] = np.nan

        temp = temperature_features["temperature_mean"].rename("temperature")
        features = temperature_features.drop(columns=["temperature_mean"])
        return temp, features

    def _merge_meter_temp(self, meter, temp):
        """
        Merge the meter and temperature dataframes and reorder the columns to have the order -
            [season, weekday_weekend, temperature, observed (if present)]

        Parameters
        ----------
            meter (pd.DataFrame): The meter dataframe.
            temp (pd.DataFrame): The temperature dataframe.

        Returns
        -------
            pd.DataFrame: The merged and transformed dataframe.
        """
        df = meter.merge(
            temp, left_index=True, right_index=True, how="left"
        ).tz_convert(meter.index.tz)
        if df["observed"].dropna().empty:
            df = df.drop(columns=["observed"])

        # Add Season and Weekday_weekend
        df["season"] = df.index.month_name().map(_const.default_season_def)
        df["weekday_weekend"] = df.index.day_name().map(
            _const.default_weekday_weekend_def
        )

        # Reorder the columns Create a list of columns
        columns = ["season", "weekday_weekend", "temperature"]
        if "observed" in df.columns:
            columns.append("observed")
        df = df[columns]

        return df

    def _check_data_sufficiency(self, sufficiency_df):
        raise NotImplementedError(
            "Can't instantiate class _DailyData, use DailyBaselineData or DailyReportingData."
        )

    def _set_data(self, data: pd.DataFrame):
        """Process data input for the Daily Model Baseline Class
        Datetime has to be either index or a separate column in the dataframe.
        Electricity data with 0 meter values are converted to NaNs.

        Parameters
        ----------
        data : pd.DataFrame
            Required columns - datetime, observed, temperature

            observed

        Returns
        -------
        processed_data : pd.DataFrame
            Dataframe appended with the correct season and day of week.
        """

        # Copy the input dataframe so that the original is not modified
        df = data.copy()

        expected_columns = [
            "observed",
            "temperature",
        ]
        # TODO maybe check datatypes
        if not set(expected_columns).issubset(set(df.columns)):
            # show the columns that are missing

            raise ValueError(
                "Data is missing required columns: {}".format(
                    set(expected_columns) - set(df.columns)
                )
            )

        # Check that the datetime index is timezone aware timestamp
        if not isinstance(df.index, pd.DatetimeIndex) and "datetime" not in df.columns:
            raise ValueError("Index is not datetime and datetime not provided")

        elif "datetime" in df.columns:
            if df["datetime"].dt.tz is None:
                raise ValueError("Datatime is missing timezone information")
            df["datetime"] = pd.to_datetime(df["datetime"])
            df.set_index("datetime", inplace=True)

        elif df.index.tz is None:
            raise ValueError("Datatime is missing timezone information")
        elif str(df.index.tz) == "UTC":
            self.warnings.append(
                EEMeterWarning(
                    qualified_name="eemeter.data_quality.utc_index",
                    description=(
                        "Datetime index is in UTC. Use tz_localize() with the local timezone to ensure correct aggregations"
                    ),
                    data={},
                )
            )
        self.tz = df.index.tz

        # prevent later issues when merging on generated datetimes, which default to ns precision
        # there is almost certainly a smoother way to accomplish this conversion, but this works
        if df.index.dtype.unit != "ns":
            utc_index = df.index.tz_convert("UTC")
            ns_index = utc_index.astype("datetime64[ns, UTC]")
            df.index = ns_index.tz_convert(self.tz)

        # Convert electricity data having 0 meter values to NaNs
        if self.is_electricity_data:
            df.loc[df["observed"] == 0, "observed"] = np.nan

        # Caltrack 2.3.2 - Drop duplicates
        df = remove_duplicates(df)

        meter = self._compute_meter_value_df(df)
        temp, temp_coverage = self._compute_temperature_features(df, meter.index)
        final_df = self._merge_meter_temp(meter, temp)
        return final_df, temp_coverage


class DailyBaselineData(_DailyData):
    """Data class to represent Daily Baseline Data.

    Only baseline data should go into the dataframe input, no blackout data should be input.
    Checks sufficiency for the data provided as input depending on OpenEEMeter specifications and populates disqualifications and warnings based on it.

    Args:
        df (DataFrame): A dataframe having a datetime index or a datetime column with the timezone also being set.
            It also requires 2 more columns - 'observed' for meter data, and 'temperature' for temperature data.
            The temperature column should have values in Fahrenheit. Please convert your temperatures accordingly.

        is_electricity_data (bool): Flag to ascertain if this is electricity data or not. Electricity data values of 0 are set to NaN.

    Attributes:
        df (DataFrame): Immutable dataframe that contains the meter and temperature values for the baseline data period.
        disqualification (list[EEMeterWarning]): A list of serious issues with the data that can degrade the quality of the model. If you want to go ahead with building the model while ignoring them, set the ignore_disqualification = True flag in the model. By default disqualifications are not ignored.
        warnings (list[EEMeterWarning]): A list of issues with the data, but none that will severely reduce the quality of the model built.

    """

    def _check_data_sufficiency(self, sufficiency_df):
        """
        Private method which checks the sufficiency of the data for daily baseline calculations using the predefined OpenEEMeter sufficiency criteria.

        Args:
            sufficiency_df (pandas.DataFrame): DataFrame containing the data for sufficiency check. Should have features such as -
            temperature_null: number of temperature null periods in each aggregation step
            temperature_not_null: number of temperature non null periods in each aggregation step

        Returns:
            disqualification (List): List of disqualifications
            warnings (list): List of warnings

        """
        # 90% coverage per period only required for billing models
        dsc = DailySufficiencyCriteria(
            data=sufficiency_df, is_electricity_data=self.is_electricity_data
        )
        dsc.check_sufficiency_baseline()
        disqualification = dsc.disqualification
        warnings = dsc.warnings

        return disqualification, warnings


class DailyReportingData(_DailyData):
    """Data class to represent Daily Reporting Data.

    Only reporting data should go into the dataframe input, no blackout data should be input.
    Checks sufficiency for the data provided as input depending on OpenEEMeter specifications and populates disqualifications and warnings based on it.

    Meter data input is optional for the reporting class.

    Args:
        df (DataFrame): A dataframe having a datetime index or a datetime column with the timezone also being set.
            It also requires 2 more columns - 'observed' for meter data, and 'temperature' for temperature data.
            The temperature column should have values in Fahrenheit. Please convert your temperatures accordingly.

        is_electricity_data (bool): Flag to ascertain if this is electricity data or not. Electricity data values of 0 are set to NaN.

    Attributes:
        df (DataFrame): Immutable dataframe that contains the meter and temperature values for the baseline data period.
        disqualification (list[EEMeterWarning]): A list of serious issues with the data that can degrade the quality of the model. If you want to go ahead with building the model while ignoring them, set the ignore_disqualification = True flag in the model. By default disqualifications are not ignored.
        warnings (list[EEMeterWarning]): A list of issues with the data, but none that will severely reduce the quality of the model built.
    """

    def __init__(self, df: pd.DataFrame, is_electricity_data: bool):
        df = df.copy()
        if "observed" not in df.columns:
            df["observed"] = np.nan

        super().__init__(df, is_electricity_data)

    @classmethod
    def from_series(
        cls,
        meter_data: pd.Series | pd.DataFrame | None,
        temperature_data: pd.Series | pd.DataFrame,
        is_electricity_data: bool | None = None,
        tzinfo: datetime.tzinfo | None = None,
    ) -> DailyReportingData:
        """Create an instance of the Data class from meter data and temperature data.

        Args:
            meter_data: The meter data to be used for the DailyReportingData instance.
            temperature_data: The temperature data to be used for the DailyReportingData instance.
            is_electricity_data: Flag indicating whether the meter data represents electricity data.
            tzinfo: Timezone information to be used for the meter data.

        Returns:
            An instance of the Data class.
        """
        if tzinfo and meter_data is not None:
            raise ValueError(
                "When passing meter data to DailyReportingData, convert its DatetimeIndex to local timezone first; `tzinfo` param should only be used in the absence of reporting meter data."
            )
        if is_electricity_data is None and meter_data is not None:
            raise ValueError(
                "Must specify is_electricity_data when passing meter data."
            )
        if meter_data is None:
            meter_data = pd.DataFrame(
                {"observed": np.nan}, index=temperature_data.index
            )
            if tzinfo:
                meter_data = meter_data.tz_convert(tzinfo)

            # If is_electricity_data is not specified, set it to True for proper functioning in the parent class. If it hits this point it's all NaNs anyway.
            if is_electricity_data is None:
                is_electricity_data = True
        if meter_data.empty:
            raise ValueError(
                "Pass meter_data=None rather than an empty series in order to explicitly create a temperature-only reporting data instance."
            )
        return super().from_series(meter_data, temperature_data, is_electricity_data)

    def _check_data_sufficiency(self, sufficiency_df):
        """
        Private method which checks the sufficiency of the data for daily reporting calculations using the predefined OpenEEMeter sufficiency criteria.

        Parameters
        ----------
        1. sufficiency_df (pandas.DataFrame): DataFrame containing the data for sufficiency check. Should have features such as -
            - temperature_null: number of temperature null periods in each aggregation step
            - temperature_not_null: number of temperature non null periods in each aggregation step

        Returns
        -------
            disqualification (List): List of disqualifications
            warnings (list): List of warnings

        """
        # 90% coverage per period only required for billing models
        dsc = DailySufficiencyCriteria(data=sufficiency_df, is_reporting_data=True)
        dsc.check_sufficiency_reporting()
        disqualification = dsc.disqualification
        warnings = dsc.warnings

        return disqualification, warnings
